/****************************************************************************
** COPYRIGHT (C) 2003-2004 ZYDAS CORPORATION                               **
** HTTP://WWW.ZYDAS.COM.TW/                                                **
****************************************************************************/

/****************************************************************************
Module Name:
    interrupt.c

This driver runs on the following hardware:
	ZD1211

Environment:
    Windows CE

Revision History

*****************************************************************************/

#include "precomp.h"

extern UCHAR     SnapBridgeTunnel[6];
extern UCHAR     Snap802_1H[6];

UCHAR
X_To_Rate(UCHAR rate, UCHAR isCCK);

#ifndef HOST_IF_USB
//-----------------------------------------------------------------------------
// Procedure:   ZD1205Isr (miniport)
//
// Description: This is the interrupt service routine.  It will check to see
//              if there are any interrupts pending, and if there are, it will
//              disable board interrupts and schedule a HandleInterrupt
//              callback.
//
// Arguments:
//  MiniportAdapterContext - The context value returned by the Miniport
//                           when the adapter was initialized (see the call
//                           NdisMSetAttributes). In reality, it is a
//                           pointer to ZD1205_ADAPTER.
//
// Result:
//  InterruptRecognized - Returns True if the interrupt belonged to this
//                        adapter, and false otherwise.
//  QueueMiniportHandleInterrupt - Returns True if we want a callback to
//                                 HandleInterrupt.
//
// Returns: (none)
//-----------------------------------------------------------------------------

VOID
ZD1205Isr(
        OUT PBOOLEAN InterruptRecognized,
        OUT PBOOLEAN QueueMiniportHandleInterrupt,
        IN NDIS_HANDLE MiniportAdapterContext
        )

{
	ULONG		EVStatus;
	PZD1205_ADAPTER Adapter;
    DEBUGFUNC(_T("ZD1205Isr"));

    Adapter = PZD1205_ADAPTER_FROM_CONTEXT_HANDLE(MiniportAdapterContext);

	// Check if any interrupt waiting for service
    ZD1205_READ_REGISTER(Adapter,InterruptCtrl, &EVStatus);

	// Keep Adapter->EnableInt synchronous with Device's status
	// Solve PC hang after "Safely remove device" problem.
	if ((EVStatus & 0x0fff0000) == (Adapter->IntMask << 16)){
		//Adapter->EnableInt = TRUE;
	}
	else{
		//Adapter->EnableInt = FALSE;
	}

	EVStatus &= Adapter->IntMask;

	// We use Adapter->IntMask to filter the interrupt belonged to the parter
	// which shared the same IRQ with us.
	// Also we use Adapter->IntMask to prevent loop call on ZD1205Isr. 
	// (In Ahu's notebook, OS will not call ZD1205Isr until (*InterruptRecognized)
	//	equals FALSE).
	// ZD1205EnableInterrupt will be invoked afetr ZD1205HandleInterrupt.
	//if (EVStatus && Adapter->EnableInt)
    {
        *InterruptRecognized = TRUE;
        *QueueMiniportHandleInterrupt = TRUE;
        ZD1205DisableInterrupt(Adapter);
    }
    //else
    {
        *InterruptRecognized = FALSE;
        *QueueMiniportHandleInterrupt = FALSE;
    }

    return;
}

//-----------------------------------------------------------------------------
// Procedure:   ZD1205HandleInterrupt
//
// Description: This routine is queued by the ISR when some interrupt
//              processing needs to be done.  It's main job is to call the
//              interrupt processing code (process receives, cleanup completed
//              transmits, etc). It will only be called with the adapter's
//              interrupts masked. This is the DPC for this driver.
//
// Arguments:
//  MiniportAdapterContext (miniport) - The context value returned by the
//                           Miniport when the adapter was initialized (see the
//                           call NdisMSetAttributes). In reality, it is a
//                           pointer to ZD1205_ADAPTER.
//
//  Context (mac) - The context value returned by the Mac when the adapter was
//            initialized. In reality, it is a pointer to ZD1205_ADAPTER.
//
// Returns: (none)
//-----------------------------------------------------------------------------

VOID
ZD1205HandleInterrupt(
                    IN NDIS_HANDLE MiniportAdapterContext
                    )

{
    ULONG      AckCommand;  
	USHORT      DPCLoopCount;
	PZD1205_ADAPTER Adapter;
	ZD1211_IRQL		OldIrql;

    DEBUGFUNC(_T("ZD1205HandleInterrupt"));

    // Context is actually our adapter
    Adapter = PZD1205_ADAPTER_FROM_CONTEXT_HANDLE(MiniportAdapterContext);

    // dpcloopcount can be changed to other values
    // to optimize for hardware/protocol timing differences
    DPCLoopCount = 2;

    //KeAcquireSpinLock(&Adapter->Lock, &OldIrql);
	ZDDbg((_T("Get SpinLock\n")));
	ZD1211_AcquireSpinLock(Adapter, &Adapter->Lock, &OldIrql);

    // Make no more than 'loopcount' loops through the interrupt processing
    // code.  We don't want to loop forever in handle interrupt if we have
    // interrupt processing to do, because this might lead to starvation
    // problems with certain protocols if we are receiving data constantly.

    while (DPCLoopCount)
    {
        // Check the interrupt status bits of the SCB, and see why we were
        // interrupted.
        //TRACE3(Adapter, ("ScbStatus %04x\n", Adapter->CSRAddress->ScbStatus));
		AckCommand = 0;
		ZD1205_READ_REGISTER(Adapter,InterruptCtrl, &AckCommand);
		AckCommand &= Adapter->IntMask;

        // If there are no interrupts to process, then exit loop
        if (!AckCommand)
            break;

        // Ack all pending interrupts now
        //Adapter->CSRAddress->ScbStatus = AckCommand;
		if (AckCommand & PS_WAKE_UP){
			ZD1205_WRITE_REGISTER(Adapter,InterruptCtrl, (PS_WAKE_UP));
			if (Adapter->DeviceInSleep){
				//++ After wake up, we should ignore all interrupt except for Wakeup Interrupt.
				//	 This is very important!
				//	 If, we do not obey this, the following bug might occurs:
				//
				//	-------------------------------------------------------------------> time
				//	^		^			^			^			^			^
				//	|		|			|			|			|			|
				//	Sleep	|			Process		Process		GetReturn	|
				//			Wakeup		WakeupInt	RxComplete	Packet		RxComplet
				//			&			(ReStart	(NotifyNdis (Rfd1)		(due to
				//			RxComplete	 Rx(Rfd1))	 Rfd1)		Chain Rfd1	 ReStartRx
				//												at last of	 in Process
				//												RfdList		 WakeupInt)
				//	This problem cause Rx-master stays in Idle 
				//	state, and we did not restart it again!
				AckCommand = 0;
				ZD1205_WRITE_REGISTER(Adapter,InterruptCtrl, (0xff));
				//--

				ProcessWakeupInterrupt(Adapter);

			}
		}

		if (AckCommand & CFG_NEXT_BCN){
			// Ack CFG_NEXT_BCN interrupt
			ZD1205_WRITE_REGISTER(Adapter,InterruptCtrl, (CFG_NEXT_BCN));
			ProcessCfgNextBcn(Adapter);
		}

		if (AckCommand & BUS_ABORT){
			// Ack BUS_ABORT interrupt
			//ZD1205_WRITE_REGISTER(Adapter,InterruptCtrl, (BUS_ABORT));
			// Go handle receive events
			// Ack ALL Interrupt
			ZD1205_WRITE_REGISTER(Adapter,InterruptCtrl, (0xff));
			ProcessBusAbortInterrupt(Adapter);
			break;
		}

		if (AckCommand & RX_COMPLETE){
			// Ack Rx_COMPLETE interrupt
			ZD1205_WRITE_REGISTER(Adapter,InterruptCtrl, (RX_COMPLETE));
			// Go handle receive events
			ProcessRXInterrupt(Adapter, &OldIrql);
		}

		if (AckCommand & TX_COMPLETE){
			// Ack TX_COMPLETE interrupt
			ZD1205_WRITE_REGISTER(Adapter,InterruptCtrl, (TX_COMPLETE));
			// Cleanup transmits
			ProcessTXInterrupt(Adapter, &OldIrql);
		}

		if (AckCommand & RETRY_FAIL){
			// Ack RETRY_FAIL interrupt
			ZD1205_WRITE_REGISTER(Adapter,InterruptCtrl, (RETRY_FAIL));
			ProcessRetryFailInterrupt(Adapter, &OldIrql);
		}


		/*
		if (AckCommand & ATIMWND_END){
			// Ack ATIMWND_END interrupt
			ZD1205_WRITE_REGISTER(Adapter,InterruptCtrl, (ATIMWND_END));
			ProcessATIMWndEndInterrupt(Adapter);
		}
		*/

        DPCLoopCount--;
    }


	//KeReleaseSpinLock(&Adapter->Lock, OldIrql);
	ZDDbg((_T("---------->>> Release SpinLock\n")));
	ZD1211_ReleaseSpinLock(Adapter, &Adapter->Lock, OldIrql);
    return;
}
#endif

//-----------------------------------------------------------------------------
// Procedure:   StartReceiveUnit
//
// Description: This routine checks the status of the ZD1205's receive unit(RU),
//              and starts the RU if it was not already active.  However,
//              before restarting the RU, the driver cleans up any recent
//              pending receives (this is very important).
//
// Arguments:
//      Adapter - ptr to Adapter object instance
//
// Returns:
//      TRUE - If we indicated any receives during this function call
//      FALSE - If we didn't indicate any receives
//-----------------------------------------------------------------------------

BOOLEAN
StartReceiveUnit(
                 IN PZD1205_ADAPTER Adapter
                 )

{
    PZD1205SwRfd      SwRfd;

    DEBUGFUNC(_T("StartReceiveUnit"));

    TRACE2(Adapter, ("\n"));

#ifdef HOST_IF_USB
	return TRUE;
#endif

    SwRfd = (PZD1205SwRfd) QueueGetHead(&Adapter->RfdList);
	ASSERT(SwRfd);
    // Set the SCB General Pointer to point the current Rfd.
	// Write high-part of addr first.
	// When low-part written, hardware will start rx automatically.
	StartUpload(Adapter, SwRfd->RfdPhys);

	return TRUE;
}

//-----------------------------------------------------------------------------
// Procedure:   checkRecPort
//
// Description: 
//-----------------------------------------------------------------------------
BOOLEAN
checkRecPort(
              IN PZD1205_ADAPTER Adapter,
			  PZD1205SwRfd       SwRfd
			  )
{
    UINT     index;
	USHORT   port;

    if ( Adapter->PERPort ){
	    index = PLCP_HEADER_SIZE + 24; // PLCP and MAC header 
        index += 34;  // get UDP port
        port = (((USHORT) SwRfd->Rfd->RxBuffer[index]) << 8) +   
			   SwRfd->Rfd->RxBuffer[index+1];
		if ( port != Adapter->PERPort )
			return FALSE; 
	}

	return TRUE;
}

//-----------------------------------------------------------------------------
// Procedure:   ProcessRXInterrupt
//
// Description: This routine will indicate any received packets to the NDIS
//              wrapper.  This routine is architected so that we will indicate
//              all of the completed receive frames that we have in our buffer,
//              before we exit the routine.  If we receive any errored frames,
//              those frames are dropped by the driver, and not indicated to
//              the ndis wrapper.
//
// Arguments:
//  Adapter - ptr to Adapter object instance
//
// Returns:
//      TRUE if a receive was indicated
//      FALSE if a receive was not indicated
//-----------------------------------------------------------------------------

BOOLEAN
ProcessRXInterrupt(
                    IN PZD1205_ADAPTER	Adapter,
                    PZD1205SwRfd        *RxSwRfdArray,
                    int                 RxSwRfdTotal,
                    ZD1211_IRQL			*pOldIrql
				   )
{
    PZD1205SwRfd      SwRfd = 0;
    PZD1205SwRfd      NextSwRfd;
    PRFD_STRUC      Rfd;
    PNDIS_PACKET    PacketArray[MAX_NUM_ALLOCATED_RFDS];
    UINT            PacketArrayCount;
    UINT            PacketFreeCount;
    UINT            i;
    BOOLEAN         ContinueToCheckRFDs=TRUE;
	PWLAN_DATA_FRM_HEADER_STRC	pWlanHdr;
	PETH_HEADER_STRUC			pEthHdr;
	UCHAR			tmp1[6];
	UCHAR			tmp2[6];
	PZD1205_DEFRAG_ENTITY		DefragEntityPtr;
	ULONG						Status;
	PNDIS_PACKET				Packet;
	PNDIS_BUFFER				Buffer;
    NDIS_STATUS AllocationStatus;
    PZD1205_DEFRAG_ENTITY *TempPtr;
	ULONG				OFFSET_SIZE;
	ULONG				OFFSET1_SIZE;
	ULONG				REMOVE_SIZE;
	ULONG				REMOVE1_SIZE;
	ULONG					typeField;
	ULONG				Count;
	ULONG				tmpvalue;
	UCHAR				Rfc1042Ptr;
	ULONG				StartCpyAddr;
	ULONG				BufferLen;
#ifdef HOST_IF_USB
    #if !fMERGE_RX_FRAME
    BOOLEAN             RunOnce = FALSE;
    #endif
#endif

	//WPA support
	PUCHAR				pByte;
	MICvar				*pMicKey;
	UCHAR				calMIC[8];
	USHORT				iv16;
	ULONG				iv32;
	UCHAR				extIV = 0;
	UCHAR				keyIndex;
	UINT				payloadLen = 1;
	UCHAR               patch;
	
	BOOLEAN				bEapol = FALSE;
#if fMERGE_RX_FRAME
    int                 SwRfdCnt;
#endif
    PWLAN_DATA_FRM_HEADER_STRC	pWlanHdr1;


    DEBUGFUNC(_T("ProcessRxInterrupt"))
    
    PacketArrayCount = 0;   
	PacketFreeCount = 0;
  
    // We stay in the while loop until we have processed all pending receives.
    for (SwRfdCnt = 0; SwRfdCnt < RxSwRfdTotal; SwRfdCnt++)
    {
        SwRfd = RxSwRfdArray[SwRfdCnt];
        // Get the pointer to our "hardware" Rfd.
        Rfd = SwRfd->Rfd;
        ASSERT(Rfd);
           
        // If the RFD does not have its complete bit set, then bail out.
        if ((SwRfd->Status = (USHORT)Rfd->CbStatus) != RFD_STATUS_COMPLETE)
        {
            break;
        }
           
        Adapter->DriverRxFrmCnt ++;
   
        
		// Get the packet length
        SwRfd->FrameLength = ((Rfd->ActualCount) & 0x3fff);
        

        //ASSERT(SwRfd->FrameLength <= MAX_WLAN_SIZE);

		Adapter->FrmRxed = TRUE;

		if (ValidateFrame(Adapter, SwRfd) == FALSE){
			Adapter->DriverDiscardedFrm ++;
			InitAndChainRfd(Adapter, SwRfd);
			continue;
		}

        // check if the address1 of the multicast frame is in the multicast list
		pWlanHdr = (PWLAN_DATA_FRM_HEADER_STRC)&SwRfd->Rfd->RxBuffer;
		if (pWlanHdr->FrameCtrl[0] != BEACON){
			Adapter->FrmRxed1 = TRUE;
		}

        if ( ( pWlanHdr->Address1[0] & 0x01 )&&( pWlanHdr->Address1[0] != 0xff ) ){
			tmpvalue = -1;
			for( i=0; i < Adapter->MulticastAddr[0]; i++){
			    tmpvalue = memcmp(&Adapter->MulticastAddr[6*i+1], pWlanHdr->Address1, 6);
				if ( tmpvalue == 0 )
					break;
			}
		    if ( tmpvalue != 0 ){
                InitAndChainRfd(Adapter, SwRfd);
				Adapter->DriverDiscardedFrmCauseByMulticastList++;
				continue;				
			}
		}

		// Matching the address2 in DefragList, if no matched address2 in List,
		// a new DefragEntity for this address2 will be linked to DefragList
		// Return FALSE, if this frame is invalid. (Sequence check)
		DefragEntityPtr = MatchingAddr2InDefragList(Adapter, SwRfd, &Status);


		if (Status != NDIS_STATUS_SUCCESS){ // This is an invalid frame
			//ASSERT(0);
			InitAndChainRfd(Adapter, SwRfd);
			Adapter->DriverDiscardedFrmCauseByFrmCtrl++;
			continue;
		}
		else{
			if ( IsMoreFrag(Adapter, SwRfd) ){
				Adapter->rxNeedFrgFrm++;
				continue;
			}
			else{ // The last fragmentation
				// filter the frame which has zero byte of the payload
				QueueRemoveEntity(&Adapter->DefragList, &DefragEntityPtr->Link);
				// Debug

				AP_Monitoring(Adapter, DefragEntityPtr);
				IBSS_Monitoring(Adapter, DefragEntityPtr);
				if ( IsDataFrame(Adapter, SwRfd) ){ // Need notice upper NDIS layer
					// if it is a OFDM frame
                    if ( SwRfd->Rfd->RxBuffer[SwRfd->Rfd->ActualCount-1] & 0x01 )
                        Adapter->CardSetting.LastSentRxRate = X_To_Rate( (UCHAR)(pWlanHdr->PlcpHdr[0] & 0x0F), 0);
					else                                      
						Adapter->CardSetting.LastSentRxRate = X_To_Rate( (UCHAR)(pWlanHdr->PlcpHdr[0] & 0x0F), 1);

					NdisAllocatePacket(&AllocationStatus,
						&Packet,
						Adapter->ReceivePacketPool);
			        ASSERT(AllocationStatus == NDIS_STATUS_SUCCESS);


					NDIS_SET_PACKET_HEADER_SIZE(Packet,
								ETHERNET_HEADER_SIZE);
					TempPtr = (PZD1205_DEFRAG_ENTITY *) 
								&Packet->MiniportReserved;
					*TempPtr = DefragEntityPtr;
					
					// Set the first NDIS buffer description
					SwRfd = (PZD1205SwRfd)QueueGetHead(
									&DefragEntityPtr->ElementList);
					ASSERT(SwRfd);
					//INITSTR((_T("HwRfd = %x\n"), SwRfd->Rfd));
					//++ Modify Header
					pWlanHdr = (PWLAN_DATA_FRM_HEADER_STRC)&SwRfd->Rfd->RxBuffer;
					// Check if MoreData bit set
					if (pWlanHdr->FrameCtrl[1] & BIT_5){
						UCHAR BroadcastAddress[ETHERNET_ADDRESS_LENGTH] = {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF};

						if (memcmp(pWlanHdr->Address1, BroadcastAddress, ETHERNET_ADDRESS_LENGTH) != 0){

							Mgt_SendNotification(Adapter, RQ_PSPOLL);
						}
					}

					//-------------------------------------------------------------------
					// |<-     C     ->|          |<-      D       ->|   REMOVE1 = C+D
					// |<- A ->|                  |<-      B       ->|   REMOVE = A+B
					// |PLCP|MAC Hdr|IV|   Data   |ICV|CRC32|EXT_INFO|
					//         |Eth Hdr|
					//         
					//	       ^       ^
					//         |       |
					//       oFFSET   OFFSET1
					//-------------------------------------------------------------------

					if(pWlanHdr->FrameCtrl[1] & 0x40){
						OFFSET_SIZE = OFFSET_SIZE_CIPHER_ON;
						OFFSET1_SIZE = OFFSET1_SIZE_CIPHER_ON;
						REMOVE_SIZE = REMOVE_SIZE_CIPHER_ON;
						REMOVE1_SIZE = REMOVE1_SIZE_CIPHER_ON;
						Rfc1042Ptr = 4;
						//WPA support
						if (((Adapter->CardSetting.EncryMode == TKIP) || (Adapter->CardSetting.EncryMode == AES)) && (Adapter->PkInstalled == 1)
							&& (pWlanHdr->IVRFC1042[3] & 0x20)){ //with extended iv bit set
							UINT len;	
											
							INITSTRWPA((_T("Extended IV bit set\n")));
							keyIndex=(pWlanHdr->IVRFC1042[3] & 0xc0) >> 6;
							INITSTRWPA((_T("keyIndex = %x\n"), keyIndex));
							extIV = 1;
							OFFSET_SIZE += EXT_IV_SIZE; //plus extended iv length
							OFFSET1_SIZE += EXT_IV_SIZE;
							REMOVE_SIZE += EXT_IV_SIZE;
							REMOVE1_SIZE += EXT_IV_SIZE;
							Rfc1042Ptr += EXT_IV_SIZE;
							iv16 = (USHORT)((pWlanHdr->IVRFC1042[0])<<8)+pWlanHdr->IVRFC1042[2];
							iv32 = *(PULONG)(&pWlanHdr->IVRFC1042[4]);
							INITSTRWPA((_T("iv16 = %x\n"), iv16));
							INITSTRWPA((_T("iv32 = %x\n"), iv32));
							
							pByte = &pWlanHdr->FrameCtrl[0];
							len = SwRfd->FrameLength-(PLCP_HEADER_SIZE+ICV_SIZE+CRC32_SIZE+EXT_INFO_SIZE);
							if (Adapter->CardSetting.EncryMode == AES){
								REMOVE_SIZE = REMOVE_SIZE - ICV_SIZE + MIC_SIZE; //CCMP no ICV field
								REMOVE1_SIZE = REMOVE1_SIZE - ICV_SIZE + MIC_SIZE;
								len += ICV_SIZE; 
								iv16 = (USHORT)((pWlanHdr->IVRFC1042[2])<<8)+pWlanHdr->IVRFC1042[0];
								iv32 = *(PULONG)(&pWlanHdr->IVRFC1042[4]);
							}	
							INITSTRWPA((_T("iv16 = %x\n"), iv16));
							INITSTRWPA((_T("iv32 = %x\n"), iv32));
							INITSTRWPA((_T("Packet Length(from Mac header to MIC) = %x\n"), len));
							INITSTRWPA((_T("Dump TKIP packet\n")));
							DumpWPA(pByte,len, FALSE, 1);
							
							if ((iv16 == Adapter->RxSeed[keyIndex].IV16) && (iv32 == Adapter->RxSeed[keyIndex].IV32)){
								INITSTRWPA((_T("iv out of sequence\n")));
								//ASSERT(0);
								NdisFreePacket(Packet);
								//KeReleaseSpinLock(&Adapter->Lock, *pOldIrql);
								ZDDbg((_T("---------->>> Release SpinLock\n")));
								ZD1211_ReleaseSpinLock(Adapter, &Adapter->Lock, *pOldIrql);
								GetReturnedMgtFrame(Adapter, DefragEntityPtr);
								//KeAcquireSpinLock(&Adapter->Lock, pOldIrql);
								ZDDbg((_T("Get SpinLock\n")));
								ZD1211_AcquireSpinLock(Adapter, &Adapter->Lock, pOldIrql);
								continue;
							}	
						}	
					}else{
						OFFSET_SIZE = OFFSET_SIZE_CIPHER_OFF;
						OFFSET1_SIZE = OFFSET1_SIZE_CIPHER_OFF;
						REMOVE_SIZE = REMOVE_SIZE_CIPHER_OFF;
						REMOVE1_SIZE = REMOVE1_SIZE_CIPHER_OFF;
						Rfc1042Ptr = 0;
					}

					patch = 0;
					if (memcmp((PVOID)&pWlanHdr->IVRFC1042[Rfc1042Ptr],SnapBridgeTunnel,6) == 0){
						patch = 1;
					}else if(memcmp((PVOID)&pWlanHdr->IVRFC1042[Rfc1042Ptr],Snap802_1H,6) == 0){
						patch = 2;
					}else if ( DefragEntityPtr->BodyLen == 0 ){
						payloadLen = 0;
					}

					//WPA support
					if ((Adapter->CardSetting.EncryMode == TKIP) && (Adapter->PkInstalled == 1) && (extIV == 1)){
						INITSTRWPA((_T("keyIndex = %x\n"), keyIndex));
						pMicKey=&Adapter->RxMicKey[keyIndex];
						MICclear(pMicKey);
						
						pByte=&pWlanHdr->Address1[0]; //DA for IBSS mode
						//if ((pWlanHdr->FrameCtrl[1] & 0x3) == 0x02)
						//	pByte=&pWlanHdr->Address3[0]; //Infrastructure mode
						
						INITSTRWPA((_T("Dump DA\n")));
						DumpWPA(pByte, 6, FALSE, 1);
						for (i=0; i<6; i++){
							MICappendByte(*pByte, pMicKey);
							pByte++;
						}
					
						pByte=&pWlanHdr->Address3[0]; //SA
						INITSTRWPA((_T("Dump SA\n")));
						DumpWPA(pByte, 6, FALSE, 1);
						for (i=0; i<6; i++){
							MICappendByte(*pByte, pMicKey);
							pByte++;
						}
					
						MICappendByte(0, pMicKey);
						MICappendByte(0, pMicKey);
						MICappendByte(0, pMicKey);
						MICappendByte(0, pMicKey);
						
					}	

					Adapter->TotalRxDataFrmBytes += DefragEntityPtr->BodyLen;
                    if ( pWlanHdr->Address1[0] & 0x01 ){
						if ( checkRecPort(Adapter, SwRfd) ){
						    Adapter->rxMulticastFrm++;
						    Adapter->rxMulticastOctets += DefragEntityPtr->BodyLen;
						}
                        if ( pWlanHdr->Address1[0] == 0xff ){
							Adapter->rxBroadcastFrm++;
							Adapter->rxBroadcastOctets += DefragEntityPtr->BodyLen;
						}
					}else{
					    Adapter->rxUnicastFrm++;
						Adapter->rxUnicastOctets += DefragEntityPtr->BodyLen;
                        Adapter->rxDataPerSec += DefragEntityPtr->BodyLen;
					}


					//Patched code for strip the RFC1042 header. 
					if ( patch == 1 ){
						if ( (Adapter->CardSetting.EncryMode == TKIP) && (Adapter->PkInstalled == 1) && (extIV == 1) ){  // if contains SNAP header
							pByte=&pWlanHdr->IVRFC1042[Rfc1042Ptr]; //SNAP header
							INITSTRWPA((_T("Dump SNAP Header\n")));
							DumpWPA(pByte, 6, FALSE, 1);
							for (i=0; i<6; i++){
								MICappendByte(*pByte, pMicKey);
								pByte++;
							}	
						}
						OFFSET_SIZE += 8;
						REMOVE_SIZE += 8;
						DefragEntityPtr->BodyLen = 
							(((USHORT)pWlanHdr->IVRFC1042[Rfc1042Ptr+6]) << 8) +
							pWlanHdr->IVRFC1042[Rfc1042Ptr+7];
					}
					else if ( patch == 2 ){
						typeField = (((USHORT)pWlanHdr->IVRFC1042[Rfc1042Ptr+6]) << 8) +
											pWlanHdr->IVRFC1042[Rfc1042Ptr+7];
                        if ((typeField != 0x8137) && (typeField != 0x80F3)) {
						    if ( (Adapter->CardSetting.EncryMode == TKIP) && (Adapter->PkInstalled == 1) && (extIV == 1) ){  // if contains SNAP header
							    pByte=&pWlanHdr->IVRFC1042[Rfc1042Ptr]; //SNAP header
							    INITSTRWPA((_T("Dump SNAP Header\n")));
							    DumpWPA(pByte, 6, FALSE, 1);
							    for (i=0; i<6; i++){
							         MICappendByte(*pByte, pMicKey);
								     pByte++;
								} 	
							}
							OFFSET_SIZE += 8;
							REMOVE_SIZE += 8;
							DefragEntityPtr->BodyLen = (USHORT)typeField;								
						}else{
						    patch = 0;
						}
						
						//if (typeField == ntohs(0x888E))
						if (typeField == 0x888E)
							bEapol = TRUE;
					}
					//End Patch


					pEthHdr = (PETH_HEADER_STRUC)((ULONG)&SwRfd->Rfd->RxBuffer +
													OFFSET_SIZE);
					for (i=0;i<ETHERNET_ADDRESS_LENGTH;i++){
						tmp1[i] = pWlanHdr->Address1[i];
						// IBSS mode -- Default
						tmp2[i] = pWlanHdr->Address2[i];
					}
					
					if ((pWlanHdr->FrameCtrl[1] & 0x3) == 0x02){
						// Infrastructure mode
						for (i=0;i<ETHERNET_ADDRESS_LENGTH;i++){
							tmp2[i] = pWlanHdr->Address3[i];
						}
					}
					else if((pWlanHdr->FrameCtrl[1] & 0x3) == 0x01){
						// AP_BSS mode
						memcpy(tmp1, pWlanHdr->Address3, 6);
						memcpy(tmp2, pWlanHdr->Address2, 6);
					}

					for (i=0;i<ETHERNET_ADDRESS_LENGTH;i++){
						pEthHdr->Destination[i] = tmp1[i];
						pEthHdr->Source[i] = tmp2[i];
					}

					pEthHdr->TypeLength[0] = (UCHAR)(DefragEntityPtr->BodyLen >> 8);
					pEthHdr->TypeLength[1] = (UCHAR)(DefragEntityPtr->BodyLen);
					//--
					NdisAllocateBuffer(&AllocationStatus,
						&Buffer,
						Adapter->ReceiveBufferPool,
						(PVOID)((ULONG)&SwRfd->Rfd->RxBuffer + OFFSET_SIZE), 
											// Only data-frame, we need to report to upper Ndis.
											// OFFSET_SIZE = PLCP_HEADER_SIZE(4/5) + 
											//				 DATA_FRM_HEADER_SIZE(24)-
											//				 ETHERNET_HEADER_SIZE(14)

						MAX_WLAN_SIZE);	
											// REMOVE_SIZE = PLCP_HEADER_SIZE(4/5) +
											//				 DATA_FRM_HEADER_SIZE(24) +
											//				 CRC32(4) +
											//				 EXT_INFO_SIZE(4) -
											//				 ETHERNET_HEADER_SIZE(14)

			        // probably should do some error handling here
			        ASSERT(AllocationStatus == NDIS_STATUS_SUCCESS);
					SwRfd->ReceiveBuffer = Buffer;
			        NdisChainBufferAtBack(Packet, Buffer);

													
					// Set the other NDIS buffer description
					StartCpyAddr = (ULONG)&SwRfd->Rfd->RxBuffer + OFFSET_SIZE +
								( SwRfd->FrameLength - REMOVE_SIZE );
					BufferLen = ( SwRfd->FrameLength - REMOVE_SIZE );

					Count = QueueGetCount(&DefragEntityPtr->ElementList);
					for(i= 0;i < Count-1;	i++){

						NextSwRfd = (PZD1205SwRfd)QueueGetNext(
									&DefragEntityPtr->ElementList, &SwRfd->Link);
						memcpy((PVOID)StartCpyAddr, 
								(PVOID)((ULONG)&NextSwRfd->Rfd->RxBuffer + OFFSET1_SIZE),
								(UINT)( NextSwRfd->FrameLength - REMOVE1_SIZE ));

						StartCpyAddr += ( NextSwRfd->FrameLength - REMOVE1_SIZE );
						BufferLen += ( NextSwRfd->FrameLength - REMOVE1_SIZE );
						QueueRemoveEntity(&DefragEntityPtr->ElementList, &NextSwRfd->Link);
						InitAndChainRfd(Adapter, NextSwRfd);
					}
					//WPA support
					if ((Adapter->CardSetting.EncryMode == TKIP) && (Adapter->PkInstalled == 1) && (extIV == 1)){
						UINT	j;
						
						
						INITSTRWPA((_T("Count = %x\n"), Count));
						
						INITSTRWPA((_T("OFFSET_SIZE = %x\n"), OFFSET_SIZE));
						INITSTRWPA((_T("REMOVE_SIZE = %x\n"), REMOVE_SIZE));
						INITSTRWPA((_T("SwRfd->FrameLength = %x\n"), SwRfd->FrameLength));
						INITSTRWPA((_T("Ndis packet length = %x\n"), BufferLen));
						INITSTRWPA((_T("Dump Ndis packet\n")));
						DumpWPA(SwRfd->Rfd->RxBuffer + OFFSET_SIZE, BufferLen, FALSE, 1);
						
						if ( payloadLen ){
							BufferLen -= MIC_SIZE;	
							if ( patch ){  // the TYPE of 802.3 packet is a part of payload 
							    pByte = SwRfd->Rfd->RxBuffer + OFFSET_SIZE + 12; //12 for DA & SA
							    payloadLen = BufferLen - 12;
							}else{ // the LENGTH of 802.3 packet is not a part of payload 							    
							    pByte = SwRfd->Rfd->RxBuffer + OFFSET_SIZE + 14; //14 for DA & SA
							    payloadLen = BufferLen - 14;
							}
							INITSTRWPA((_T("Payload Length = %x\n"), payloadLen));
							INITSTRWPA((_T("Dump Payload\n")));
							DumpWPA(pByte, payloadLen, FALSE, 1);
							
							if (bEapol)
							{
								INITSTRWPA((_T("Rx a EAPOL packet\n")));
								//EapolParser(Adapter, pByte, payloadLen);
							}
												
							for (j=0; j<payloadLen; j++){ //payload not include SNAP header
								MICappendByte(*pByte, pMicKey);
								pByte++;
							}
						}

						MICgetMIC(calMIC, pMicKey);
						INITSTRWPA((_T("Dump Calculate MIC\n")));
						DumpWPA(calMIC, 8, FALSE, 1);
						
						pByte = (PUCHAR)StartCpyAddr - MIC_SIZE; //point to MIC field
						INITSTRWPA((_T("Dump Receive MIC\n")));
						DumpWPA(pByte, 8, FALSE, 1);
						
						if (memcmp(calMIC, pByte, MIC_SIZE)!=0){
							INITSTR((_T("#####MIC error!!!!#####\n"))); //TBD
							//KeReleaseSpinLock(&Adapter->Lock, Adapter->OldIrql);
							//ZD1205GetReturnedPackets(Adapter, PacketArray[PacketArrayCount]);
							//KeAcquireSpinLock(&Adapter->Lock, &Adapter->OldIrql);

						
							//MIC BUG
							//Mgt_SendNotification(Adapter, RQ_TKIPDBG);	
							ASSERT(0);

							if (Adapter->keyMgrState == 0){
								if (pWlanHdr->Address1[0] & 1)
								{
									Adapter->WPAMICErrorLog[0]=1;
									NdisMoveMemory(Adapter->WPAMICErrorLog+2,pWlanHdr->Address2,6);
									ReportAuthEvent(Adapter,  NDIS_802_11_AUTH_REQUEST_GROUP_ERROR);
								} else {
									Adapter->WPAMICErrorLog[1]=1;
									NdisMoveMemory(Adapter->WPAMICErrorLog+8,pWlanHdr->Address2,6);
									ReportAuthEvent(Adapter,  NDIS_802_11_AUTH_REQUEST_PAIRWISE_ERROR);
								}
								
#ifdef DISALLOW_DUP_REPORT_MIC_FAIL
								Adapter->keyMgrState = 1;
#endif								
							}
							//continue;
						}	
						else
							INITSTRWPA((_T("******MIC Success!!!!******\n")));

						//update iv sequence
						Adapter->RxSeed[keyIndex].IV16 = iv16;
						Adapter->RxSeed[keyIndex].IV32 = iv32;
					}	

					NdisAdjustBufferLength((PNDIS_BUFFER) Buffer,
						(UINT) BufferLen);
					
					// Process WLAN Forwarding
					if (Adapter->CardSetting.BSSType == AP_BSS){
						if (pEthHdr->Destination[0] & BIT_0){ // Multicast
							AP_ForwardMulticastPkt(Adapter, Packet, pEthHdr);
						}
						else{	// Unicast
							if (AP_CheckForward(Adapter, pEthHdr)){
								if(!AP_ForwardUnicastPkt(Adapter, Packet, pEthHdr)){
									//KeReleaseSpinLock(&Adapter->Lock, *pOldIrql);
									ZDDbg((_T("---------->>> Release SpinLock\n")));
									ZD1211_ReleaseSpinLock(Adapter, &Adapter->Lock, *pOldIrql);
									ZD1205GetReturnedPackets(Adapter, Packet);
									//KeAcquireSpinLock(&Adapter->Lock, pOldIrql);
									ZDDbg((_T("Get SpinLock\n")));
									ZD1211_AcquireSpinLock(Adapter, &Adapter->Lock, pOldIrql);
								}
								continue;
							}
						}
					}
                    
					PacketArray[PacketArrayCount] = Packet;

					// set the status on the packet, either resources or success
					if (QueueGetCount(&Adapter->RfdList) > MIN_NUM_RFD)
					{
						NDIS_SET_PACKET_STATUS(PacketArray[PacketArrayCount],NDIS_STATUS_SUCCESS);
						// Keep track of the highest array index in which a packet
						// with status NDIS_STATUS_SUCCESS is found. Once we hit the low
						// resource condition, all further packets will be marked with status
						// NDIS_STATUS_RESOURCES. This is because we hold the adapter lock
						// here which prevents UsedRfdCount/NumRfd from being modified
						// elsewhere.
					}
					else
					{
						NDIS_SET_PACKET_STATUS(PacketArray[PacketArrayCount],
							NDIS_STATUS_RESOURCES);
                        
						PacketFreeCount++;
						
					}

					PacketArrayCount++;
								
				}
				else
				{
					//ASSERT(0);
					Adapter->DriverRxMgtFrmCnt++;

					//Fixed Lost first EAPOL Key packet in WPA-PSK,
					//because we drop data packet when Adapter not ready. 
					//Marvell will resend seconed EAPOL Key packet after 2 seconeds,
					//but it often cause timeout - Roger Huang 2004/06/15
                    if (Adapter->Infra_Process_Associate)
					{
			           PWLAN_ASSOCIATE_RSP_FRM_STRUC  pAssRspFrm;
			           pAssRspFrm = (PWLAN_ASSOCIATE_RSP_FRM_STRUC)&SwRfd->Rfd->RxBuffer;
			           if ( (pAssRspFrm->FrameCtrl[0] & 0xfc) == 0x10 ){ // Association response frame
				         if ((pAssRspFrm->Status[0] == 0) &&
				             (pAssRspFrm->Status[1] == 0) ){
  			                 INITSTR((_T("Association Response\n")));
				             Adapter->adapterReady=TRUE;
				             Adapter->AID = (((USHORT)(pAssRspFrm->AID[1] & 0x3f)) << 8) +
							 pAssRspFrm->AID[0];
				             INITSTR((_T("Our AID is %x\n"), Adapter->AID));
				             Adapter->PkInstalled = 0; // 3808 by yarco

							if (Adapter->MoreConnect == 0){
						       NdisMIndicateStatus(Adapter->ZD1205AdapterHandle,
						                    NDIS_STATUS_MEDIA_CONNECT,
						                    NULL,     // Not used.
						                    0);                 // Not used.
					            NdisMIndicateStatusComplete(Adapter->ZD1205AdapterHandle);
							 }
						 }	
					   }
					}

					if ( Adapter->CardSetting.BSSType == AP_BSS )
					{
						pWlanHdr1 = (PWLAN_DATA_FRM_HEADER_STRC)&SwRfd->Rfd->RxBuffer;

						if ( pWlanHdr->FrameCtrl[0] == AUTHENTICATE )
						{
							AP_ProcessAuthen(Adapter, DefragEntityPtr);
							AP_GetReturnedMgtFrame(Adapter, DefragEntityPtr);
						}
						else if ( pWlanHdr->FrameCtrl[0] == ASSOCIATE )
						{
							AP_ProcessAssociationReq(Adapter, DefragEntityPtr);
							AP_GetReturnedMgtFrame(Adapter, DefragEntityPtr);
						}
				        else if ( pWlanHdr->FrameCtrl[0] == REASSOCIATE_REQ )
						{
				            AP_ProcessReAssociationReq(Adapter, DefragEntityPtr);
							AP_GetReturnedMgtFrame(Adapter, DefragEntityPtr);
						}
						else if (QueueGetCount(&Adapter->RfdList) <= 5)
						{
							USHORT    Qi;
							PZD1205SwRfd  QSwRfd;
							ULONG    QCount;

							QCount = QueueGetCount(&DefragEntityPtr->ElementList);
							for (Qi=0;Qi < QCount;Qi++)
							{
								QSwRfd = (PZD1205SwRfd)QueuePopHead(&DefragEntityPtr->ElementList);
								InitAndChainRfd(Adapter, QSwRfd);
							}
 
							QueuePutTail(&Adapter->FreeDefragEntityList, &DefragEntityPtr->Link);
						}     
						else
						{                      
							Mgt_SendQueue(Adapter, DefragEntityPtr);
						}
					}
					else
					{
						
						if ( (QueueGetCount(&Adapter->RfdList) <= 5) )
						{
							USHORT    Qi;
							PZD1205SwRfd  QSwRfd;
							ULONG    QCount;

							QCount = QueueGetCount(&DefragEntityPtr->ElementList);
							for (Qi=0;Qi < QCount;Qi++)
							{
								QSwRfd = (PZD1205SwRfd)QueuePopHead(&DefragEntityPtr->ElementList);
								InitAndChainRfd(Adapter, QSwRfd);
							}
 
							QueuePutTail(&Adapter->FreeDefragEntityList, &DefragEntityPtr->Link);
						}     
						else
						{    
							Mgt_SendQueue(Adapter, DefragEntityPtr);
						}					
					}					
				}
			}
		}            
    }// end while 1
        
    // if we didn't process any receives, just return from here
    if (PacketArrayCount){
		Adapter->Traffic += PacketArrayCount;
		//KeReleaseSpinLock(&Adapter->Lock, *pOldIrql);
		ZDDbg((_T("---------->>> Release SpinLock\n")));
		ZD1211_ReleaseSpinLock(Adapter, &Adapter->Lock, *pOldIrql);
        NdisMIndicateReceivePacket(Adapter->ZD1205AdapterHandle,
			PacketArray,
            PacketArrayCount);
		//KeAcquireSpinLock(&Adapter->Lock, pOldIrql);
		ZDDbg((_T("Get SpinLock\n")));
		ZD1211_AcquireSpinLock(Adapter, &Adapter->Lock, pOldIrql);
		Adapter->NotifyNDISRxFrmCnt += (USHORT)PacketArrayCount;
		Adapter->GoodReceives += PacketArrayCount;
 		Adapter->DataTrafficLight = TRUE;
           
    }
  
	if (PacketFreeCount){
		//KeReleaseSpinLock(&Adapter->Lock, *pOldIrql);
		ZDDbg((_T("---------->>> Release SpinLock\n")));
		ZD1211_ReleaseSpinLock(Adapter, &Adapter->Lock, *pOldIrql);
		for (i=PacketArrayCount-PacketFreeCount;i<PacketArrayCount ;i++ )
		{
			ZD1205GetReturnedPackets(Adapter, PacketArray[i]);
		} // end for 
		//KeAcquireSpinLock(&Adapter->Lock, pOldIrql);
		ZDDbg((_T("Get SpinLock\n")));
		ZD1211_AcquireSpinLock(Adapter, &Adapter->Lock, pOldIrql);
		PerformAgingFunction(Adapter);
	}

    return (TRUE);
}

//-----------------------------------------------------------------------------
// Procedure:   ProcessTXInterrupt
//
// Description: This routine is called by HandleInterrupt to process transmit
//              interrupts for ZD1205 adapters.  Basically, this routine will
//              remove any completed packets from the active transmit chain,
//              append them to the completed list.  TransmitCleanup will be
//              called to free up the TCBs, map registers, etc, that the
//              packets on the completed list consumed.  At the end of this
//              routine, the MAC version of this driver will try to send any
//              packets that it queued because of an earlier lack of resources.
//
// Arguments:
//      Adapter - ptr to Adapter object instance
//
// Return:
//      TRUE - If we indicated any loopback packets during this function call
//      FALSE - If we didn't indicate any loopaback packets
//-----------------------------------------------------------------------------

BOOLEAN
ProcessTXInterrupt(PZD1205_ADAPTER Adapter,
                   ZD1211_IRQL     *pOldIrql
				   )
{
    PZD1205SwTcb        SwTcb;
	PZD1205SwTcb		NextSwTcb;
    BOOLEAN             Status = FALSE;
    PNDIS_PACKET        Packet;
    PNDIS_BUFFER        CurrBuff;
#ifndef HOST_IF_USB
    INT                 MapRegToFree;
#endif
	UINT					PacketLength;
	ULONG				DebugPrint;
	USHORT				i;
	PSTA_CONTEXT		STAContextPtr;
#ifdef HOST_IF_USB
	BOOLEAN				RunOnce = FALSE;
#endif
	UCHAR               aid;
	PVOID               FirstBufferVA;
	UINT                FirstBufferLength, TotalBufferLength;


    DEBUGFUNC(_T("ProcessTXInterrupt"));

	DebugPrint = Adapter->DbgMsgFlag & DBGMSG_TX;
    if (QueueEmpty(&Adapter->ActiveChainList))
    {
        return FALSE;
    }


    // Look at the TCB at the head of the queue.  If it has been completed
    // then pop it off and place it at the tail of the completed list.
    // Repeat this process until all the completed TCBs have been moved to the
    // completed list
    while (SwTcb = (PZD1205SwTcb) QueueGetHead(&Adapter->ActiveChainList))
    {
#ifdef HOST_IF_USB
		// in USB modem, only run once
		if (RunOnce)	break;
		RunOnce = TRUE;
#endif
        // check to see if the TCB has been DMA'd
		// Workaround for hardware problem that seems leap over a TCB
		// and then fill completion token in the next TCB.
        if (SwTcb->Tcb->CbStatus != CB_STATUS_COMPLETE){
			NextSwTcb = SwTcb;
			while(1){
				NextSwTcb = (PZD1205SwTcb)QueueGetNext(&Adapter->ActiveChainList, 
																&NextSwTcb->Link);
				if (NextSwTcb){
					if (NextSwTcb->Tcb->CbStatus == CB_STATUS_COMPLETE){
						break;
					}
				}
				else{
					break;
				}
			}
			if (!NextSwTcb)
				break;
		}

        // Remove the TCB from the active queue.
        SwTcb = (PZD1205SwTcb) QueuePopHead(&Adapter->ActiveChainList);

		if (Adapter->CardSetting.BSSType == AP_BSS){
			STAContextPtr = (PSTA_CONTEXT)SwTcb->STAContextPtr;
			if (STAContextPtr){
				// JSWAN 7/27/2004
				// Minus here, added at retry fail
				//STAContextPtr->IdleTime = 0;
				if (STAContextPtr->IdleTime > 0){
					STAContextPtr->IdleTime --;
				}
                aid = STAContextPtr->AID;
                Adapter->OppositeRateInfo[aid].ConsecutiveSuccessFrames++;
			}
		}      
 		else{
           for( i=0; i<Adapter->OppositeCnt; i++ ){
				if ( !memcmp(&SwTcb->HwHeaderPtr->MacHeader[4], Adapter->OppositeRateInfo[i].MacAddr, 6) ){
							
					Adapter->OppositeRateInfo[i].ConsecutiveSuccessFrames++;

					break;
				}
			}
		}

		TransmitCleanup(Adapter, SwTcb);

		if (!SwTcb->LastFrag){
			continue;
		}


		// It's the last fragmentation.
		Packet = SwTcb->Packet;
		if (!SwTcb->Coalesce){
			//NdisQueryPacket(Packet, NULL, NULL, &SwTcb->FirstBuffer, NULL);
			NdisGetFirstBufferFromPacket( Packet, 
				                          &SwTcb->FirstBuffer, 
										  &FirstBufferVA, 
										  &FirstBufferLength, 
										  &TotalBufferLength
										  );
			CurrBuff = SwTcb->FirstBuffer;
			// free the map register associated with each buffer
		}else{
			// Free coalesce buffer
            QueuePutTail(&Adapter->CoalesceBufferList, &SwTcb->Coalesce->Link);
		}


		if (SwTcb->Provider == UPPER_NDIS){
		//if (SwTcb->NoticeNDIS){
			//KeReleaseSpinLock(&Adapter->Lock, *pOldIrql);
			//ZDDbg((_T("---------->>> Release SpinLock\n")));
			//ZD1211_ReleaseSpinLock(Adapter, &Adapter->Lock, *pOldIrql);
			// Do a send Complete for this frame
			Adapter->GoodTransmits ++;
#ifdef CE_ALL
			Adapter->TxTimer = GetTickCount();
#else
			Adapter->TxTimer = KeQueryInterruptTime();
#endif


#ifdef HOST_IF_USB
			if (!SwTcb->bHasCompleteBeforeSend){
				//NdisMSendComplete(
				//	Adapter->ZD1205AdapterHandle,
				//	Packet,
				//	NDIS_STATUS_SUCCESS);
			}
#else
			NdisMSendComplete(
				Adapter->ZD1205AdapterHandle,
				Packet,
				NDIS_STATUS_SUCCESS);
#endif
			//KeAcquireSpinLock(&Adapter->Lock, pOldIrql);
			//ZDDbg((_T("Get SpinLock\n")));
			//ZD1211_AcquireSpinLock(Adapter, &Adapter->Lock, pOldIrql);

			Adapter->SequenceNum  ++;
			Adapter->DataTrafficLight = TRUE;
		}
		else{
			Adapter->EnetFrameLock = FALSE;
			//ASSERT(0);
			if ( SwTcb->Provider != AP_FORWARD_MUL )
			    //NdisQueryPacket(Packet, NULL, NULL, &SwTcb->FirstBuffer, &PacketLength);
				NdisGetFirstBufferFromPacket( Packet, 
											  &SwTcb->FirstBuffer, 
											  &FirstBufferVA, 
											  &FirstBufferLength, 
											  &PacketLength
											  );
            if (SwTcb->FrameType == MGT_LBTEST)
            {
            //JWEI 2004/01/29
            #if !fZD1211_LOOPBACK
                Adapter->LBTxCnt += PacketLength;
                Mgt_SendNotification(Adapter, RQ_LBTEST);
            #else
                Adapter->LBTxCnt ++;
                if (Adapter->LBTxCnt >= cLBTEST_COUNT)
                    Adapter->Notification &= ~RQ_LBTEST;
                else
                    Mgt_SendNotification(Adapter, RQ_LBTEST);
            #endif
			}
			ProviderMSendComplete(Adapter, SwTcb, pOldIrql);
		}

	}

#ifdef HOST_IF_USB

	if (!QueueEmpty(&Adapter->ActiveChainList))
	{
		//DbgPrint(" SetEvent StartTxEvent from ProcessTXInterrupt\n");
		//SetEvent(Adapter->StartTxEvent);		
	}

	//USB_MPSenToTxEP(	Adapter,
	//					Adapter->pTxUrb,
	//					Adapter->pTxIrp,
	//					Adapter->pTxContext
	//					);
#endif	

    // If we queued any transmits because we didn't have any TCBs earlier,
    // dequeue and send those packets now, as long as we have free TCBs.
	//ProcessQueuedPackets(Adapter);


    return TRUE;
}


//-----------------------------------------------------------------------------
// Procedure:   ProcessRetryFailInterrupt
//
// Description: 
//
// Arguments:
//      Adapter - ptr to Adapter object instance
//
// Return:
//
//-----------------------------------------------------------------------------
BOOLEAN
ProcessRetryFailInterrupt(PZD1205_ADAPTER Adapter, ZD1211_IRQL *pOldIrql)
{
    PZD1205SwTcb        SwTcb;
	PZD1205SwTcb		NextSwTcb;
	PHWTCB_STRUC		HwTcb;
    BOOLEAN             Status = FALSE;
    PNDIS_PACKET        Packet;
    PNDIS_BUFFER        CurrBuff;
#ifndef HOST_IF_USB
    INT                 MapRegToFree;
#endif
	ULONG				tmpvalue;
	USHORT				Len;
	UCHAR				CurrentRate;
	UCHAR				NewRate;
	USHORT				NextLen;
	USHORT				LenInUs;
	USHORT				NextLenInUs;
	UCHAR				Service;
	PCTRL_STRUC			pCtrlSet;
	UCHAR				ShortPreambleFg;
	PVOID               FirstBufferVA;
	UINT                FirstBufferLength, TotalBufferLength;

    DEBUGFUNC(_T("ProcessRetryFailInterrupt"));

    if (QueueEmpty(&Adapter->ActiveChainList))
    {
		INITSTR((_T("Check if Tx bus master is in Retry-Fail-Wait state\n")));
		ZD1205_READ_REGISTER(Adapter,DeviceState, &tmpvalue);
		ASSERT( (tmpvalue & 0xf) == 0xa );
		// Solve the FTP-long-time-test hang problem.
		// Re-Start Tx-Bus master with an suspend TCB
		SwTcb = (PZD1205SwTcb) QueueGetHead(&Adapter->TxCBList);
		StartDownload(Adapter, SwTcb->TcbPhys | BIT_0);

        return FALSE;
    }

	// Feature: Rate Adaption
	// - During the procedure of processing a transmitting frame, we must keep
	//   the TaRate consistent.
	// - When to fall OppositeContext.CurrentTxRate:
	//   Whenever RetryFail occurs, change OppositeContext.CurrentTxRate by a value
	//   ((Rate of this TCB) minus a degree) and modify this TCB's control-setting 
	//   with the OppositeContext.CurrentTxRate and then Restart this TCB.
	//	 (Set RetryMAX = 2).
	//   Once the TxRate is 1M and still RetryFail, abandon this frame.
	// - When to rise TxRate:
	//   If there are 10 frames transmitted successfully 
	//   (OppositeContext.ConsecutiveSuccessFrames >= 10), change 
	//   OppositeContext.CurrentTxRate by a value
	//   ((Rate of this TCB) plus a degree).
	// - Adjust OppositeContext.CurrentTxRate manually. (by application tool)

	SwTcb = (PZD1205SwTcb) QueueGetHead(&Adapter->ActiveChainList);
	pCtrlSet = SwTcb->HwCtrlPtr;

	CurrentRate = (pCtrlSet->CtrlSetting[0] & 0x1f);

	if (CurrentRate > 3){
		switch (CurrentRate){
			case 0x1b:	// 6M
				CurrentRate = 4;
				break;
			case 0x1f:	// 9M
				CurrentRate = 5;
				break;
			case 0x1a:	// 12M
				CurrentRate = 6;
				break;
			case 0x1e:	// 18M
				CurrentRate = 7;
				break;
			case 0x19:	// 24M
				CurrentRate = 8;
				break;
			case 0x1d:	// 36M
				CurrentRate = 9;
				break;
			case 0x18:	// 48M
				CurrentRate = 0xa;
				break;
			case 0x1c:	// 54M
				CurrentRate = 0xb;
				break;
		}
	}

	ShortPreambleFg = (pCtrlSet->CtrlSetting[0] & 0x20);

	if ( ((!ShortPreambleFg) && (CurrentRate > 0) && (!Adapter->RateLock)) ||
		 ((ShortPreambleFg) && (CurrentRate > 1) && (!Adapter->RateLock)) ){ 
		// Fall TxRate a degree
		NewRate = Rate_Adaption(Adapter, &pCtrlSet->CtrlSetting[12], CurrentRate, FALL_RATE);
		
		// LenInUs, Service
		Len = (pCtrlSet->CtrlSetting[1] + ((USHORT)pCtrlSet->CtrlSetting[2] << 8));
		Cal_Us_Service(NewRate, Len, &LenInUs, &Service);
		pCtrlSet->CtrlSetting[20] = (UCHAR)LenInUs;
		pCtrlSet->CtrlSetting[21] = (UCHAR)(LenInUs >> 8);
		pCtrlSet->CtrlSetting[22] = Service;
		// NextLenInUs
#ifdef OFDM
		NextLen = (pCtrlSet->CtrlSetting[CTRL_SIZE+1] + ((USHORT)pCtrlSet->CtrlSetting[CTRL_SIZE+2] << 8));
#else
		NextLen = (pCtrlSet->CtrlSetting[18] + ((USHORT)pCtrlSet->CtrlSetting[19] << 8));
#endif
		Cal_Us_Service(NewRate, NextLen, &NextLenInUs, &Service);
		pCtrlSet->CtrlSetting[23] = (UCHAR)NextLenInUs;
		pCtrlSet->CtrlSetting[24] = (UCHAR)(NextLenInUs >> 8);

		if (NewRate > 3){
			switch (NewRate){
				case 0x04:	// 6M
					NewRate = 0x1b;
					break;
				case 0x05:	// 9M
					NewRate = 0x1f;
					break;
				case 0x06:	// 12M
					NewRate = 0x1a;
					break;
				case 0x07:	// 18M
					NewRate = 0x1e;
					break;
				case 0x08:	// 24M
					NewRate = 0x19;
					break;
				case 0x09:	// 36M
					NewRate = 0x1d;
					break;
				case 0x0a:	// 48M
					NewRate = 0x18;
					break;
				case 0x0b:	// 54M
					NewRate = 0x1c;
					break;
				default:
					ASSERT(0);
					break;
			}
		}

		// Modify Control-setting
		pCtrlSet->CtrlSetting[0] = (ShortPreambleFg | NewRate);
		pCtrlSet->CtrlSetting[11] |= BIT_0; // Set need backoff

		INITSTR((_T("Check if Tx bus master is in Retry-Fail-Wait state\n")));
		ZD1205_READ_REGISTER(Adapter,DeviceState, &tmpvalue);
		ASSERT( (tmpvalue & 0xf) == 0xa );
		// Re-Start Tx-Bus master with a lower rate
		StartDownload(Adapter, SwTcb->TcbPhys | BIT_0);
		return TRUE;
	}


    while (SwTcb = (PZD1205SwTcb) QueueGetHead(&Adapter->ActiveChainList))
    {
        SwTcb = (PZD1205SwTcb) QueuePopHead(&Adapter->ActiveChainList);
		TransmitCleanup(Adapter, SwTcb);
        // check to see if the TCB has been DMA'd
        if (!SwTcb->LastFrag)
			continue;

		Adapter->RetryFailCnt ++;

		// It's the last fragmentation.
		Packet = SwTcb->Packet;
		if (!SwTcb->Coalesce){
			//NdisQueryPacket(Packet, NULL, NULL, &SwTcb->FirstBuffer, NULL);
			NdisGetFirstBufferFromPacket( Packet, 
										  &SwTcb->FirstBuffer, 
										  &FirstBufferVA, 
										  &FirstBufferLength, 
										  &TotalBufferLength
										  );
			CurrBuff = SwTcb->FirstBuffer;
			// free the map register associated with each buffer
			while (CurrBuff)
			{
				// free a map register
#ifndef HOST_IF_USB
				MapRegToFree = Adapter->OldestUsedMapReg;
				Adapter->OldestUsedMapReg++;
				if (Adapter->OldestUsedMapReg == Adapter->NumMapRegisters)
				Adapter->OldestUsedMapReg = 0;
		
				// Release the map register and its associated physical mapping
				NdisMCompleteBufferPhysicalMapping(
					Adapter->ZD1205AdapterHandle,
					CurrBuff,
					MapRegToFree);
#endif

				// Get the next buffer that needs the be "de-mapped"
				NdisGetNextBuffer(CurrBuff, &CurrBuff);
			}
		}else{
			// Free coalesce buffer
			QueuePutTail(&Adapter->CoalesceBufferList, &SwTcb->Coalesce->Link);
		}

		//Mark2o04_1++ This block should be in the front of the Mark2o04_2
		//			   to prevent from preempted by ZD1205MultipleSend.
		//			   That's because Mark2o04_2 will release Adapter->Lock
		//			   Temporarily, so the ZD1205MultipleSend has the chance
		//			   to be called.
		NextSwTcb = (PZD1205SwTcb) QueueGetHead(&Adapter->ActiveChainList);
		INITSTR((_T("Check if Tx bus master is in Retry-Fail-Wait state\n")));
		ZD1205_READ_REGISTER(Adapter,DeviceState, &tmpvalue);
		ASSERT( (tmpvalue & 0xf) == 0xa );
		if (NextSwTcb){
			// Re-Start Tx bus master
			// Set BIT_0 to escape from Retry-Fail-Wait state.
			StartDownload(Adapter, NextSwTcb->TcbPhys | BIT_0);
		}
		else{
			// Re-Start Tx-Bus master with an suspend TCB
			HwTcb = (PHWTCB_STRUC)SwTcb->Tcb;
			// Set BIT_0 to escape from Retry-Fail-Wait State.
			StartDownload(Adapter, HwTcb->NextCbPhyAddrLowPart | BIT_0);
		}
		//Mark2o04_1--

		//Mark2o04_2++
		if (SwTcb->Provider==UPPER_NDIS){
			//KeReleaseSpinLock(&Adapter->Lock, *pOldIrql);
			ZDDbg((_T("---------->>> Release SpinLock\n")));
			ZD1211_ReleaseSpinLock(Adapter, &Adapter->Lock, *pOldIrql);
			// Do a send Complete for this frame
			if (Adapter->CardSetting.EncryMode != TKIP){
				NdisMSendComplete(
					Adapter->ZD1205AdapterHandle,
					Packet,
					NDIS_STATUS_SUCCESS);
					//NDIS_STATUS_FAILURE);
			}
			else{
				NdisMSendComplete(
					Adapter->ZD1205AdapterHandle,
					Packet,
					//NDIS_STATUS_SUCCESS);
					NDIS_STATUS_FAILURE);
			}
			//KeAcquireSpinLock(&Adapter->Lock, pOldIrql);
			ZDDbg((_T("Get SpinLock\n")));
			ZD1211_AcquireSpinLock(Adapter, &Adapter->Lock, pOldIrql);
		}
		else{
			ProviderMSendComplete(Adapter, SwTcb, pOldIrql);
		}
		//Mark2o04_2--
		break;
    }

	Adapter->IBSS_Wakeup_Dest = TRUE;

    // If we queued any transmits because we didn't have any TCBs earlier,
    // dequeue and send those packets now, as long as we have free TCBs.
	ProcessQueuedPackets(Adapter);

	return TRUE;
}

//-----------------------------------------------------------------------------
// Procedure:   ProcessATIMWndEndInterrupt
//
// Description: 
//
// Arguments:
//      Adapter - ptr to Adapter object instance
//
// Return:
//
//-----------------------------------------------------------------------------

BOOLEAN
ProcessATIMWndEndInterrupt(PZD1205_ADAPTER Adapter)
{
	ULONG	tmpvalue;
	ULONGLONG			TSFTimer;
    DEBUGFUNC(_T("ProcessATIMWndEndInterrupt"));

	ZD1205_READ_REGISTER(Adapter,TSF_LowPart, &tmpvalue);
	TSFTimer = tmpvalue;
	ZD1205_READ_REGISTER(Adapter,TSF_HighPart, &tmpvalue);
	TSFTimer += (((ULONGLONG)tmpvalue) << 32);
	TSFTimer = TSFTimer >> 10; // in unit of TU
	INITSTR((_T("TSF(TU) %d \n"), TSFTimer));
	INITSTR((_T("BeaconInterval = %d\n"), Adapter->CardSetting.BeaconInterval));
	INITSTR((_T("TSF mod BeaconInterval = %d\n"), (TSFTimer % Adapter->CardSetting.BeaconInterval)));

	if (Adapter->PwrState && (Adapter->adapterReady) &&
		(!Adapter->DeviceInSleep) &&
		(Adapter->CardSetting.BSSType == INDEPENDENT_BSS)){
		Mgt_SendNotification(Adapter, RQ_ATIM_WND_END);
	}
	return TRUE;
}

//-----------------------------------------------------------------------------
// Procedure:   ProcessCfgNextBcn
//
// Description: 
//
// Arguments:
//      Adapter - ptr to Adapter object instance
//
// Return:
//
//-----------------------------------------------------------------------------

BOOLEAN
ProcessCfgNextBcn(PZD1205_ADAPTER Adapter)
{

    NDIS_STATUS         Status;
    PNDIS_PACKET        QueuePacket;

    DEBUGFUNC(_T("ProcessCfgNextBcn"));

	if (Adapter->CardSetting.BSSType == INDEPENDENT_BSS){
		Adapter->FrmRxed1 = FALSE;
	}
	else if (Adapter->CardSetting.BSSType == AP_BSS){
		if (Adapter->mDtimCount == 0){
			Adapter->mDtimCount = (UCHAR)Adapter->CardSetting.DTIMPeriod;
			while (Adapter->FirstMulPendingQueue && Adapter->bcstAnnounce){
				//Adapter->QueueCount--;
				QueuePacket = Adapter->FirstMulPendingQueue;
				DequeuePacket( Adapter->FirstMulPendingQueue, 
							   Adapter->LastMulPendingQueue
							   );	
				Adapter->NumMulPendingQueued--;
				if (Adapter->FirstTxQueue){
					Adapter->NumPacketsQueued++;
					EnqueuePacket( Adapter->FirstTxQueue, 
						           Adapter->LastTxQueue, 
						           QueuePacket
								   );
				}
				else{
					//Status = NDIS_GET_PACKET_STATUS(QueuePacket);
					//if ( Status != NDIS_STATUS_SUCCESS ){
					    Status = SetupNextSend( Adapter, 
						                        QueuePacket, 
							    			    DATA, 
						                        Adapter->CardSetting.EncryMode, 
									    	    UPPER_NDIS, 
										        NULL
										        );
					/*}else{
					    Status = SetupNextSend( Adapter, 
						                        QueuePacket, 
							    			    DATA, 
						                        Adapter->CardSetting.EncryMode, 
									    	    AP_FORWARD_UNI, 
										        NULL
										        );
					}*/
					
					if (Status == NDIS_STATUS_RESOURCES)
					{
						Adapter->NumPacketsQueued++;
						EnqueuePacket( Adapter->FirstTxQueue, 
							           Adapter->LastTxQueue, 
							           QueuePacket
									   );
					}
				}
			}
		}
		Adapter->mDtimCount -= 1;
		Mgt_SendNotification(Adapter, RQ_AP_UPDATEBEACON);	
	}


	return TRUE;
}

// Matching the address2 in DefragList, if no matched address2 in List,
// a new DefragEntity for this address2 will be linked to DefragList
// Return FALSE, if this frame is invalid. (Sequence check)
PZD1205_DEFRAG_ENTITY
MatchingAddr2InDefragList(PZD1205_ADAPTER Adapter, 
							PZD1205SwRfd SwRfd,
							PULONG pStatus)
{
	PZD1205_DEFRAG_ENTITY	DefragEntityPtr;
	PZD1205SwRfd			tmpRfd;
	PWLAN_DATA_FRM_HEADER_STRC	pWlanHdr1;
	PWLAN_DATA_FRM_HEADER_STRC	pWlanHdr2;
	USHORT					i;
	USHORT					j;
	ULONG					Count;
	ULONG					REMOVE1_SIZE;

	DefragEntityPtr = (PZD1205_DEFRAG_ENTITY)QueueGetHead(&Adapter->DefragList);
	pWlanHdr2 = (PWLAN_DATA_FRM_HEADER_STRC)&SwRfd->Rfd->RxBuffer;
	if(pWlanHdr2->FrameCtrl[1] & 0x40){
		REMOVE1_SIZE = ENCAP_LEN_CIPHER_ON;
					// REMOVE1_SIZE = PLCP_HEADER_SIZE(4/5) +
					//				 DATA_FRM_HEADER_SIZE(24) +
					//				 IV(4) +
					//				 ICV(4) +
					//				 CRC32(4) +
					//				 EXT_INFO_SIZE(4)
		if (pWlanHdr2->IVRFC1042[3] & 0x20){
			if (Adapter->CardSetting.EncryMode == AES){
				REMOVE1_SIZE = REMOVE1_SIZE - ICV_SIZE + MIC_SIZE;
			}	
		}				
	}else{
		REMOVE1_SIZE = ENCAP_LEN_CIPHER_OFF;
					// REMOVE1_SIZE = PLCP_HEADER_SIZE(4/5) +
					//				 DATA_FRM_HEADER_SIZE(24) +
					//				 CRC32(4) +
					//				 EXT_INFO_SIZE(4)
	}
	// Search whole the DefragList to find a DefragEntity with address2 the same
	// with the incoming frame.
	Count = QueueGetCount(&Adapter->DefragList);
	for (i=0; i < Count;i++ ){
		tmpRfd = (PZD1205SwRfd)QueueGetTail(&DefragEntityPtr->ElementList);
		pWlanHdr1 = (PWLAN_DATA_FRM_HEADER_STRC)&tmpRfd->Rfd->RxBuffer;
		for (j=0;j<6;j++){
			if (pWlanHdr1->Address2[j] != pWlanHdr2->Address2[j]){
				break;
			}
		}

		if (j == 6){ // Found
			// Need to check the sequence-control of this frame
			if ( (pWlanHdr2->SeqCtrl[1] == pWlanHdr1->SeqCtrl[1]) &&
					(pWlanHdr2->SeqCtrl[0] == (pWlanHdr1->SeqCtrl[0]+1)) ){
				DefragEntityPtr->BodyLen += (USHORT)(SwRfd->Rfd->ActualCount - REMOVE1_SIZE);
				QueuePutTail(&DefragEntityPtr->ElementList, &SwRfd->Link);
				*pStatus = NDIS_STATUS_SUCCESS;
				return(DefragEntityPtr);
			}
			else if( (pWlanHdr2->SeqCtrl[1] == pWlanHdr1->SeqCtrl[1]) &&
					 (pWlanHdr2->SeqCtrl[0] == pWlanHdr1->SeqCtrl[0]) ){
				// This frame is a duplication frame
				*pStatus = NDIS_STATUS_FAILURE;
				return(NULL);
			}
			else{ // Sequence number has been out-of-order, reset this DefragEntity
				while(!QueueEmpty(&DefragEntityPtr->ElementList)){
					tmpRfd = (PZD1205SwRfd)QueuePopHead(&DefragEntityPtr->ElementList);
					InitAndChainRfd(Adapter, tmpRfd);
				}
				QueueRemoveEntity(&Adapter->DefragList, &DefragEntityPtr->Link);
				QueuePutTail(&Adapter->FreeDefragEntityList, &DefragEntityPtr->Link);
				// Now, go to NewGuy
				goto NewGuy;
			}
		}
		else{
			DefragEntityPtr = (PZD1205_DEFRAG_ENTITY)QueueGetNext(
								&Adapter->DefragList, &DefragEntityPtr->Link);
		}
	}

NewGuy:
	// Could not find it.

	// Check if FragNo == 0
	if (pWlanHdr2->SeqCtrl[0] & 0xf){ // Invalid
		*pStatus = NDIS_STATUS_FAILURE;
		return(NULL);
	}

	// Check if duplication frame detected.
	if (Adapter->CardSetting.BSSType != AP_BSS){
		if ((pWlanHdr2->SeqCtrl[0] == Adapter->LatestSeqCtrl[0]) &&
			(pWlanHdr2->SeqCtrl[1] == Adapter->LatestSeqCtrl[1])){
			if ( !Adapter->notCheckSN ){
				*pStatus = NDIS_STATUS_FAILURE;
				return(NULL);
			}
		}
		
		// Save the latest SeqCtrl.
		Adapter->LatestSeqCtrl[0] = pWlanHdr2->SeqCtrl[0];
		Adapter->LatestSeqCtrl[1] = pWlanHdr2->SeqCtrl[1];
	}

	// Create a DefragEntity for it.
	DefragEntityPtr = (PZD1205_DEFRAG_ENTITY)QueuePopHead(
								&Adapter->FreeDefragEntityList);
	ASSERT(DefragEntityPtr);
	DefragEntityPtr->BodyLen = (USHORT)(SwRfd->Rfd->ActualCount - REMOVE1_SIZE);
	QueuePutTail(&Adapter->DefragList, &DefragEntityPtr->Link);
	QueuePutTail(&DefragEntityPtr->ElementList, &SwRfd->Link);
	*pStatus = NDIS_STATUS_SUCCESS;
	return(DefragEntityPtr);
}

void
PerformAgingFunction(PZD1205_ADAPTER Adapter)
{
	PZD1205_DEFRAG_ENTITY	DefragEntityPtr;
	PZD1205SwRfd			SwRfd;
    DEBUGFUNC(_T("PerformAgingFunction"));

	INITSTR((_T("\n")));
	//ASSERT(0);
	while ( ! QueueEmpty(&Adapter->DefragList) ){
		DefragEntityPtr = (PZD1205_DEFRAG_ENTITY)QueuePopHead(&Adapter->DefragList);
		while (! QueueEmpty(&DefragEntityPtr->ElementList)){
			SwRfd = (PZD1205SwRfd)QueuePopHead(&DefragEntityPtr->ElementList)
			INITSTR((_T("HwRfd = %x\n"), SwRfd->Rfd));
			InitAndChainRfd(Adapter, SwRfd);
		}
		QueuePutTail(&Adapter->FreeDefragEntityList, &DefragEntityPtr->Link);
	}
	return;
}

UCHAR	BroadcastBSSID[6] = {0xff,0xff,0xff,0xff,0xff,0xff};

BOOLEAN
ValidateFrame(PZD1205_ADAPTER Adapter, 
			  PZD1205SwRfd      SwRfd
			  )
{
	PWLAN_DATA_FRM_HEADER_STRC	pWlanHdr;
	ULONG	i;
    NDIS_STATUS     Status;
	PSTA_CONTEXT	STAContextPtr;
	PNDIS_PACKET    QueuePacket;

	// Accept Data/Management frame only.
	pWlanHdr = (PWLAN_DATA_FRM_HEADER_STRC)&SwRfd->Rfd->RxBuffer;

    if (SwRfd->FrameLength == 0)
        return FALSE;
	
	//SwRfd->Rfd->RxBuffer[SwRfd->Rfd->ActualCount-1]
	//bit7:	error frame
	//bit6:	crc16 error
	//bit5:	address not match
	//bit4: crc32 error
	//bit3:	decrypt error
	//bit2: overrun
	//bit1: Rx Timeout
	//bit0:	OFDM modulation

#ifdef HOST_IF_USB
	if ( SwRfd->Rfd->RxBuffer[SwRfd->Rfd->ActualCount-1] & 0x80 )
	{
		return FALSE;
	}
	
	// this may be HMAC bug
	if ((Adapter->PkInstalled) && 
			IsDataFrame(Adapter, SwRfd) && 
			(pWlanHdr->FrameCtrl[1] & 0x40)){
			
		ZDDbgPrint((_T("Decrypt status 0x%x\n"), SwRfd->Rfd->RxBuffer[SwRfd->Rfd->ActualCount-2]));
		if ((SwRfd->Rfd->RxBuffer[SwRfd->Rfd->ActualCount-2] != 2) &&
			(Adapter->CardSetting.EncryMode == TKIP))
		{	
			// error, pairwise key has install, but recive data not encrypt
			ZDDbgPrint((_T("Pairwise key has install, but recive data decrypt wrong!!\n")));
			return FALSE;
		}
	}
#endif

	//if ( (pWlanHdr->FrameCtrl[0] & 0xc) == DATA )
	//	return FALSE;

	if (Adapter->TestFlag & DUMP_RX_FRAME){
		ZDDbgPrint((_T("Dump Frame ......\n")));
		for (i=0;i<SwRfd->FrameLength;i++){
			ZDDbgPrint((_T("%2x "), SwRfd->Rfd->RxBuffer[i]));
		}
		ZDDbgPrint((_T("End\n")));
	}

	if (pWlanHdr->FrameCtrl[0] == MGT_SCANTEST)
		return TRUE;

	if ((!Adapter->GkInstalled)&&
		((Adapter->CardSetting.AuthMode == Ndis802_11AuthModeWPA)||
		 (Adapter->CardSetting.AuthMode == Ndis802_11AuthModeWPAPSK))&&
		((Adapter->CardSetting.EncryMode == TKIP)||(Adapter->CardSetting.EncryMode == AES))&&
		(pWlanHdr->Address1[0] & 0x01)&&
		((pWlanHdr->FrameCtrl[0] & 0xc) == DATA)){

		ZDDbgPrint((_T("Encrypt BroadCast frame but Group Key setting not OK!!\n")));

		return FALSE;
	}

	// Check if the SwRfd->FrameLength matched the length derived from PLCP.
#if !(defined(GCCK) && defined(OFDM))
	PlcpUs = pWlanHdr->PlcpHdr[2]+(((USHORT)pWlanHdr->PlcpHdr[3]) << 8);
	switch(pWlanHdr->PlcpHdr[0]){
		case 0x0A:
			MacLength = (PlcpUs >> 3);
			break;
		case 0x14:
			MacLength = (PlcpUs >> 2);
			break;
		case 0x37:
			MacLength = ((PlcpUs*11) >> 4);
			break;
		case 0x6E:
			MacLength = ((PlcpUs*11) >> 3);
			if (pWlanHdr->PlcpHdr[1] & BIT_7){
				MacLength -=1;
			}
			break;
		case 0xa5:	//16.5M
			MacLength = (((PlcpUs*33) >> 4) - 
						2*((pWlanHdr->PlcpHdr[1] & BIT_6) >> 6) -
						((pWlanHdr->PlcpHdr[1] & BIT_7) >> 7));
			break;
		case 0xdc:	//22M
			MacLength = (((PlcpUs*11) >> 2) - 
						2*((pWlanHdr->PlcpHdr[1] & BIT_6) >> 6) -
						((pWlanHdr->PlcpHdr[1] & BIT_7) >> 7));
			break;
		case 0x13:	//27.5M
			MacLength = (((PlcpUs*55) >> 4) - 
						2*((pWlanHdr->PlcpHdr[1] & BIT_6) >> 6) -
						((pWlanHdr->PlcpHdr[1] & BIT_7) >> 7));
			break;			
		case 0x4a:	//33M
			MacLength = (((PlcpUs*33) >> 3) - 
						4*((pWlanHdr->PlcpHdr[1] & BIT_5) >> 5) -
						2*((pWlanHdr->PlcpHdr[1] & BIT_6) >> 6) -
						((pWlanHdr->PlcpHdr[1] & BIT_7) >> 7));
			break;
		case 0x81:	//38.5M
			MacLength = (((PlcpUs*77) >> 4) - 
						4*((pWlanHdr->PlcpHdr[1] & BIT_5) >> 5) -
						2*((pWlanHdr->PlcpHdr[1] & BIT_6) >> 6) -
						((pWlanHdr->PlcpHdr[1] & BIT_7) >> 7));
			break;
		case 0xb8:	//44M
			MacLength = (((PlcpUs*11) >> 1) - 
						4*((pWlanHdr->PlcpHdr[1] & BIT_5) >> 5) -
						2*((pWlanHdr->PlcpHdr[1] & BIT_6) >> 6) -
						((pWlanHdr->PlcpHdr[1] & BIT_7) >> 7));
			break;
		case 0xef:	//49.5M
			MacLength = (((PlcpUs*99) >> 4) - 
						4*((pWlanHdr->PlcpHdr[1] & BIT_5) >> 5) -
						2*((pWlanHdr->PlcpHdr[1] & BIT_6) >> 6) -
						((pWlanHdr->PlcpHdr[1] & BIT_7) >> 7));
			break;
		case 0x26:	//55M
			MacLength = (((PlcpUs*55) >> 3) - 
						4*((pWlanHdr->PlcpHdr[1] & BIT_5) >> 5) -
						2*((pWlanHdr->PlcpHdr[1] & BIT_6) >> 6) -
						((pWlanHdr->PlcpHdr[1] & BIT_7) >> 7));
			break;
		case 0x5d:	//60.5M
			MacLength = (((PlcpUs*121) >> 4) - 
						4*((pWlanHdr->PlcpHdr[1] & BIT_5) >> 5) -
						2*((pWlanHdr->PlcpHdr[1] & BIT_6) >> 6) -
						((pWlanHdr->PlcpHdr[1] & BIT_7) >> 7));
			break;
		case 0x53:	//8.25M
			MacLength = (((PlcpUs*33) / 32) - 
						2*((pWlanHdr->PlcpHdr[1] & BIT_6) >> 6) -
						((pWlanHdr->PlcpHdr[1] & BIT_7) >> 7));
			break;
		default:
			//ASSERT(0);
			Adapter->PLCP_MacLen_Mismatch ++;
			return FALSE;
			break;
	}
	if (MacLength != (SwRfd->FrameLength - PLCP_EXTINFO_SIZE)){	// Minus PLCP(4)+EXT_INFO(4)
		
		//ZD1205_WRITE_REGISTER(Adapter,0x804, 0x200);
		//ZD1205_READ_REGISTER(Adapter,0x808, &tmpvalue);
		//for (i=0;i<tmpvalue;i++){
		//	ZD1205_READ_REGISTER(Adapter,0x800, &tmpvalue1);
		//}
		//ASSERT(0);
		
		Adapter->PLCP_MacLen_Mismatch ++;
		return FALSE;
	}
#endif

#if !(defined(GCCK) && defined(OFDM))
	// for 3Com AP which 
    if ( (Adapter->CheckCTS == 1)&& 
		 ((pWlanHdr->FrameCtrl[0] & 0xfc) == 0xC4) ){
		ZD1205_READ_REGISTER(Adapter,CtlReg1, &tmpvalue);
		Adapter->CheckCTS = 0;
		if ( pWlanHdr->PlcpHdr[0] == 0x0A ){
			tmpvalue &= 0xE3;
		    tmpvalue |= 0x04;
		    ZD1205_WRITE_REGISTER(Adapter,CtlReg1, tmpvalue);
			ZD1205_WRITE_REGISTER(Adapter,Ack_Timeout_Ext, 0x3f);	
		}else if ( pWlanHdr->PlcpHdr[0] == 0x14 ){
			if ( (tmpvalue & 0x1C) != 0x04 ){
				tmpvalue &= 0xE3;
				tmpvalue |= 0x04;
		        ZD1205_WRITE_REGISTER(Adapter,CtlReg1, tmpvalue);
			}
		}else if ( pWlanHdr->PlcpHdr[0] == 0x37 ){
			if ( (tmpvalue & 0x1C) != 0x08 ){
				tmpvalue &= 0xE3;
				tmpvalue |= 0x08;
		        ZD1205_WRITE_REGISTER(Adapter,CtlReg1, tmpvalue);
			}
		}else if ( pWlanHdr->PlcpHdr[0] == 0x6E ){
			if ( (tmpvalue & 0x1C) != 0x0C ){
				tmpvalue &= 0xE3;
				tmpvalue |= 0x0C;
		        ZD1205_WRITE_REGISTER(Adapter,CtlReg1, tmpvalue);
			}
		}
	}
#endif

	if (Adapter->CardSetting.BSSType == AP_BSS){
		STAContextPtr = FindSTAContext(Adapter, pWlanHdr->Address2);
		if ((pWlanHdr->FrameCtrl[0] & 0xc) == DATA){
			if (!STAContextPtr){
				if (!(pWlanHdr->Address1[0] & BIT_0)){	// Unicast
					AP_SendDisAssociation(Adapter, pWlanHdr->Address2, 
						REASON_NONAUTHEN);
				}
				return FALSE;
			}
			if (STAContextPtr->LinkStatus != CONNECTED){
				if (!(pWlanHdr->Address1[0] & BIT_0)){	// Unicast
					AP_SendDisAssociation(Adapter, pWlanHdr->Address2, 
						REASON_NONASSOC);
				}
				return FALSE;
			}
			STAContextPtr->IdleTime = 0;
			// Monitor the change of STA power state.
			if (pWlanHdr->FrameCtrl[1] & BIT_4){
				if (!STAContextPtr->PSMode){
					ZDDbgPrint((_T("STA enter power-saving mode\n")));
					STAContextPtr->PSMode = TRUE;
					Adapter->aSTAInPSMode = TRUE;
				}
			}
			else{
				if (STAContextPtr->PSMode){
					while(STAContextPtr->FirstPendingQueue){
				        //STAContextPtr->UniQCount--;
						QueuePacket = STAContextPtr->FirstPendingQueue;
						DequeuePacket( STAContextPtr->FirstPendingQueue, 
							           STAContextPtr->LastPendingQueue
									   );	
						STAContextPtr->NumPendingQueued--;
						if (Adapter->FirstTxQueue){
							Adapter->NumPacketsQueued++;							
							EnqueuePacket( Adapter->FirstTxQueue, 
								           Adapter->LastTxQueue, 
								           QueuePacket
										   );
						}
						else{
							Status = NDIS_GET_PACKET_STATUS(QueuePacket);
							if ( Status != NDIS_STATUS_SUCCESS ){
							    Status = SetupNextSend( Adapter, 
								                        QueuePacket, 
									    			    DATA, 
								                        Adapter->CardSetting.EncryMode, 
											    	    UPPER_NDIS, 
												        NULL
												        );
							}else{
							    Status = SetupNextSend( Adapter, 
								                        QueuePacket, 
									    			    DATA, 
								                        Adapter->CardSetting.EncryMode, 
											    	    AP_FORWARD_UNI, 
												        NULL
												        );
							}
							if (Status == NDIS_STATUS_RESOURCES){
								Adapter->NumPacketsQueued++;
							    EnqueuePacket( Adapter->FirstTxQueue, 
								               Adapter->LastTxQueue, 
								               QueuePacket
										       );
							}
						}
					}
					STAContextPtr->PSMode = FALSE;
					if (!AP_AnySTAInPSMode(Adapter)){
						Adapter->aSTAInPSMode = FALSE;
					}
				}
			}
		}
		else if(pWlanHdr->FrameCtrl[0] == PS_POLL){
			if (!STAContextPtr){
				return FALSE;
			}
			if (STAContextPtr->LinkStatus != CONNECTED){
				return FALSE;
			}
			if (STAContextPtr->FirstPendingQueue){
				ZDDbgPrint((_T("Get a PS_POLL\n")));
				ZDDbgPrint((_T("MAC = %x %x %x %x %x %x\n"), 
					               STAContextPtr->MacAddr[0], 
								   STAContextPtr->MacAddr[1],
								   STAContextPtr->MacAddr[2],
								   STAContextPtr->MacAddr[3],
								   STAContextPtr->MacAddr[4],
								   STAContextPtr->MacAddr[5]
								   ));
				//STAContextPtr->UniQCount--;
				STAContextPtr->PolledBySTA = TRUE;
				QueuePacket = STAContextPtr->FirstPendingQueue;
				DequeuePacket( STAContextPtr->FirstPendingQueue, 
							   STAContextPtr->LastPendingQueue
							   );	
				STAContextPtr->NumPendingQueued--;
				if (Adapter->FirstTxQueue){
					Adapter->NumPacketsQueued++;
					EnqueuePacket( Adapter->FirstTxQueue, 
						           Adapter->LastTxQueue, 
						           QueuePacket
								   );
				}
				else{
					Status = NDIS_GET_PACKET_STATUS(QueuePacket);
					if ( Status != NDIS_STATUS_SUCCESS ){
					    Status = SetupNextSend( Adapter, 
					                            QueuePacket, 
							    			    DATA, 
						                        Adapter->CardSetting.EncryMode, 
									    	    UPPER_NDIS, 
										        NULL
										        );
					}else{
					    Status = SetupNextSend( Adapter, 
						                        QueuePacket, 
							    			    DATA, 
						                        Adapter->CardSetting.EncryMode, 
									    	    AP_FORWARD_UNI, 
										        NULL
										        );
					}
					
					if (Status == NDIS_STATUS_RESOURCES){
						Adapter->NumPacketsQueued++;
						EnqueuePacket( Adapter->FirstTxQueue, 
							           Adapter->LastTxQueue, 
							           QueuePacket);
					}
				}
			}
		}
	}

	if(pWlanHdr->FrameCtrl[1] & 0x40){
		// Minimum Length = PLCP(4/5)+MACHeader(24)+IV(4)+EXTINFO(4)+ICV(4)+CRC(4)
		if (SwRfd->FrameLength <= ENCAP_LEN_CIPHER_ON){
			if ((SwRfd->FrameLength == ENCAP_LEN_CIPHER_ON) &&
				((pWlanHdr->FrameCtrl[0] & 0xc) == MANAGEMENT)){
				// Filter zero data packet to prevent page-fault in some systems.
				// Pass zero management frame for testing.
			}
			else{
				Adapter->LessThanDataMinLen ++;
				return FALSE;
			}
		}
	}
	else{
		// Minimum Length = PLCP(4/5)+MACHeader(24)+EXTINFO(4)+CRC(4)
		if (SwRfd->FrameLength <= ENCAP_LEN_CIPHER_OFF){
			if ((SwRfd->FrameLength == ENCAP_LEN_CIPHER_OFF) &&
				((pWlanHdr->FrameCtrl[0] & 0xc) == MANAGEMENT)){
				// Filter zero data packet to prevent page-fault in some systems.
				// Pass zero management frame for testing.
			}
			else{
				Adapter->LessThanDataMinLen ++;
				return FALSE;
			}
		}
	}

	// Check if FrameLength > MAX_WLAN_SIZE.
	if (SwRfd->FrameLength > MAX_WLAN_SIZE){
		// Do not worry about the corruption of HwRfd.
		// If the FrameLength > 2410, the Rx-Bus-Master skip the bytes that exceed
		// 2410 to protect the structure of HwRfd. 
		// However, the Rx-Bus-Master still reports this frame to host if the frame 
		// is recognized as good by the FA(Frame Analyzer).
		//ASSERT(0);
		Adapter->GreaterThanMaxLen ++;
		return FALSE;
	}

	//Not inidicate data frame to NDIS when adapter not ready
	//Add by Roger Huang 2004.1.5 
    if ((!Adapter->adapterReady) && ((pWlanHdr->FrameCtrl[0] & 0xc) == DATA)) {
	   return FALSE;
	}

	// calculate Signal quality and strength
	if (Adapter->adapterReady && (pWlanHdr->Address1[0] & BIT_0)){ //multicast
		if ((pWlanHdr->FrameCtrl[0] & 0xc) == MANAGEMENT){ // Mgt frame
			if (memcmp(pWlanHdr->Address3, Adapter->BSSID, 6) != 0){
				// not the BSSID we connect
				
				if ((pWlanHdr->FrameCtrl[0] == PROBEREQUEST) ||
					(pWlanHdr->FrameCtrl[0] == PROBERESPONSE) ||
					(pWlanHdr->FrameCtrl[0] == MGT_LBTEST) ||
					(pWlanHdr->FrameCtrl[0] == MGT_SCANTEST)){
					
					return TRUE;
				}
				
				if ((pWlanHdr->FrameCtrl[0] == BEACON) &&
					((Adapter->CardSetting.BSSType == INDEPENDENT_BSS)||
					 (Adapter->CardSetting.BSSType == AP_BSS))){
					
					return TRUE;
				}

				return FALSE;
			}
		}
		else if ((pWlanHdr->FrameCtrl[0] & 0xc) == DATA){ // Data frame
			if (Adapter->CardSetting.BSSType == INFRASTRUCTURE_BSS){
				if (memcmp(pWlanHdr->Address2, Adapter->BSSID, 6) != 0){
					// not the BSSID we connect
					return FALSE;
				}
			}
			else if (Adapter->CardSetting.BSSType == INDEPENDENT_BSS){
				if (memcmp(pWlanHdr->Address3, Adapter->BSSID, 6) != 0){
					// not the BSSID we connect
					return FALSE;
				}
			}
			// if now is pseduo IBSS pass all
		}
	}
	else if (!Adapter->adapterReady){
		// not connect yet, no need to calculate strngth
		return TRUE;
	}

	{
        Adapter->RxSignalQuality1 = SwRfd->Rfd->RxBuffer[SwRfd->Rfd->ActualCount-EXT_INFO_SIZE];
        Adapter->RxSignalQuality2 = SwRfd->Rfd->RxBuffer[SwRfd->Rfd->ActualCount-EXT_INFO_SIZE + 2];

		Adapter->RxSignalQuality = CalculateQuality(Adapter, SwRfd, &Adapter->RxSignalQualityIndB);
		Adapter->RxSignalStrength = CalculateStrength(Adapter, SwRfd);
#ifdef HOST_IF_USB
        if (SwRfd->Rfd->RxBuffer[SwRfd->Rfd->ActualCount-1] & 0x1)
			Adapter->PHYFreOFDMframe = 1;
#endif
		
    //``JWEI 2003/12/22
    #if fANT_DIVERSITY || fTX_PWR_CTRL
        if (!(SwRfd->Rfd->RxBuffer[SwRfd->Rfd->ActualCount-1] & 0x01))
    #endif
        {                                   // CCK, not OFDM modulation
            Adapter->Acc_Num ++;
			if (Adapter->bUseSignalQuality4AntDiv==0)
				Adapter->Acc_SQ += Adapter->RxSignalStrength;
			else
				Adapter->Acc_SQ += Adapter->RxSignalQuality;
        //    JWeiDbgPrint2((" <1:%d, %d> ", Adapter->Acc_SQ, Adapter->Acc_Num));
        #if fTX_PWR_CTRL
        //    Adapter->TxPwrCCK ++;
        #endif
        }
    #if fANT_DIVERSITY || fTX_PWR_CTRL
        else
        {
            Adapter->Acc_Num_OFDM ++;

			if (Adapter->bUseSignalQuality4AntDiv==0)
				Adapter->Acc_SQ_OFDM += Adapter->RxSignalStrength;
			else
				Adapter->Acc_SQ_OFDM += Adapter->RxSignalQuality;
        //    JWeiDbgPrint2((" <2:%d, %d> ", Adapter->Acc_SQ_OFDM, Adapter->Acc_Num_OFDM));
        #if fTX_PWR_CTRL
        // Rx OFDM ==> Ack 24M
            if (Adapter->TxOFDMType != cTX_OFDM)
                Adapter->TxOFDMCnt = 0;
            Adapter->TxOFDMType = cTX_OFDM;
        //    Adapter->TxPwrOFDM ++;
        #endif
        }
    #endif
    }

	return TRUE;
}

#if 0
UCHAR SQ1_Constant[] =
{
    0x0E, 0x10, 0x12, 0x15, 0x17, 0x19, 0x1D, 0x20,
    0x24, 0x28, 0x2D, 0x32, 0x39, 0x40, 0x48, 0x51,
    0x5A, 0x65, 0x71, 0x80, 0x8F, 0xA1, 0xB5,
};
#endif

ULONG X_Constant[] = {
					715,
					655,
					585,
					540,
					470,
					410,
					360,
					315,
					270,
					235,
					205,
					175,
					150,
					125,
					105,
					85,
					65,
					50,
					40,
					25,
					15
};

UCHAR
X_To_Rate(UCHAR rate, UCHAR isCCK)
{
    if ( isCCK )
	{
	    switch(rate)
		{
		    case 0x0A: return 0x00;  // 1M
			case 0x14: return 0x01;  // 2M
			case 0x37: return 0x02;  // 5.5M
			case 0x6E: return 0x03;  // 11M
		}		
	}
    else
	{
	    switch(rate)
		{
		    case 0x0B: return 0x04;   // 6M
			case 0x0F: return 0x05;   // 9M
			case 0x0A: return 0x06;   // 12M
			case 0x0E: return 0x07;   // 18M
			case 0x09: return 0x08;   // 24M
			case 0x0D: return 0x09;   // 36M
			case 0x08: return 0x0A;   // 48M
			case 0x0C: return 0x0B;   // 54M
		}
	
	}

	return 0x03;
}

UCHAR
X_To_dB(ULONG X, UCHAR rate)
{
	UCHAR	ret = 0;
	int		i;

//``JWEI 2003/12/22
    switch (rate)
    {
        case 0x0B:  // 6M
        case 0x0A:  // 12M
        case 0x09:  // 24M
            X /= 2;
            break;
        case 0x0F:  // 9M
        case 0x0E:  // 18M
        case 0x0D:  // 36M
        case 0x0C:  // 54M
            X *= 3;
            X /= 4;
            break;
        case 0x08:  // 48M
            X *= 2;
            X /= 3;
            break;
        default:
            break;
    }

    for (i = 0; i < sizeof(X_Constant); i++)
    {
        if (X > X_Constant[i])
            break;
    }
    
    switch (rate)
    {
        case 0x0B:  // 6M
        case 0x0F:  // 9M
            ret = i + 3;
            break;
        case 0x0A:  // 12M
        case 0x0E:  // 18M
            ret = i + 5;
            break;
        case 0x09:  // 24M
        case 0x0D:  // 36M
            ret = i + 9;
            break;
        case 0x08:  // 48M
        case 0x0C:  // 54M
		ret = i + 15;
		break;
	default:
		break;
	}

	return ret;
}

UCHAR
CalculateQuality(	PZD1205_ADAPTER Adapter,
					PZD1205SwRfd	SwRfd,
					PUCHAR			pQualityIndB
					)
{
	// return in %
	// the value from PHY is in 

	UCHAR						CorrectQuality = 0;
	PWLAN_DATA_FRM_HEADER_STRC	pWlanHdr;
	
	UCHAR                       SignalQuality2 = Adapter->RxSignalQuality2;
	UINT						Len = SwRfd->Rfd->ActualCount - EXT_INFO_SIZE;
	ULONG						X;
	USHORT						tmpf;
//double                      tmpf;

	pWlanHdr = (PWLAN_DATA_FRM_HEADER_STRC)&SwRfd->Rfd->RxBuffer;

    if (SwRfd->Rfd->RxBuffer[SwRfd->Rfd->ActualCount-1] & 0x01)
    {
		// it's OFDM
		Adapter->rxOFDMDataFrame++;
		
    //``JWEI 2003/12/23
        X = 10000 * SignalQuality2 / Len;
        CorrectQuality = X_To_dB(X, (UCHAR)(pWlanHdr->PlcpHdr[0] & 0x0F));
		
		if(pQualityIndB)
			*pQualityIndB = CorrectQuality;
        
		JWeiDbgPrint2((" [SQ2:%d, %d] ", SignalQuality2, CorrectQuality));
        CorrectQuality = CorrectQuality * 4;
        if (CorrectQuality > 100)
            CorrectQuality = 100;
    }
    else
    {
        // it's CCK
        Adapter->rx11bDataFrame++;
        
        // the value from PHY is in scale from Max is 0 and Min is 0xb5
        switch(pWlanHdr->PlcpHdr[0])
        {
            case 0x0A:  // 1 M 
            case 0x14:  // 2 M
            case 0x37:  // 5.5 M
            case 0x6E:  // 11 M
            //``JWEI 2003/12/23
				tmpf = 0;

			    if (Adapter->RxSignalQuality1 > 0)
                    tmpf = (USHORT)(ZDLog10multiply100(Adapter->RxSignalQuality1) * 20 /100);
		  
				CorrectQuality = 45 - (UCHAR) (tmpf);
               
				if(pQualityIndB)
					*pQualityIndB = CorrectQuality;
                
				CorrectQuality = CorrectQuality * 4;
                if (CorrectQuality > 100)
                    CorrectQuality = 100;

				break;
			default:
				break;
		}
	}

	return CorrectQuality;
}

UCHAR
CalculateStrength(	PZD1205_ADAPTER Adapter,
					PZD1205SwRfd	SwRfd
					)
{
	// return in ? , the Value-105 = dB
	// the value from PHY is in ?
    UCHAR i, rssi, tmp;
    ULONG tmpvalue=2;
	PWLAN_DATA_FRM_HEADER_STRC	pWlanHdr;

	pWlanHdr = (PWLAN_DATA_FRM_HEADER_STRC)&SwRfd->Rfd->RxBuffer;

	rssi = SwRfd->Rfd->RxBuffer[SwRfd->Rfd->ActualCount-EXT_INFO_SIZE+1];

	if ( (((Adapter->CardSetting.BSSType == INFRASTRUCTURE_BSS)&&
		   (!memcmp(pWlanHdr->Address2, Adapter->BSSID, 6)))||
		  ((Adapter->CardSetting.BSSType == INDEPENDENT_BSS)&&
		   (!memcmp(pWlanHdr->Address3, Adapter->BSSID, 6)))||
		  (Adapter->CardSetting.BSSType == PSEUDO_IBSS))&&
		 (Adapter->adapterReady) )
	{
		for(i=0; i<Adapter->PHYTestIndex-1; i++)
			tmpvalue *= 2; 

		if ( (Adapter->CR122Flag == 1)||(Adapter->CR203Flag == 1) )
			rssi += 22;
		tmp = Adapter->PHYTestRssi;
		Adapter->PHYTestTotal = Adapter->PHYTestTotal 
								- (Adapter->PHYTestTotal/tmpvalue)
			    				+ rssi;

		Adapter->PHYTestRssi = (UCHAR) (Adapter->PHYTestTotal/tmpvalue);
		Adapter->FrameCountDuring3Sec++;
		Adapter->FrameCountDuring10ms++;
    }
	//return SwRfd->Rfd->RxBuffer[SwRfd->Rfd->ActualCount-EXT_INFO_SIZE+1];
#if JXIAODBG
	if (Adapter->AntDivState==ACC_2 &&
		Adapter->RssiListIndex <= 255)
	{//jxiaoDbg
		ULONG	systemUpTime;
		USHORT	lwSystemUpTime;
		NdisGetSystemUpTime(&systemUpTime);
		lwSystemUpTime=(USHORT)systemUpTime;
		Adapter->RssiListDuringMT2[Adapter->RssiListIndex & 255]=
			rssi + (lwSystemUpTime << 16);
		// LowWord: Rssi, Highword: recvdTime
		Adapter->RssiListIndex++;
		//DbgPrint("RecvTime:%u: RSSI = %u , ANT=%u\n",
		//	systemUpTime,
		//	rssi,
		//	Adapter->CurrentAntenna);
	}
#endif
	return rssi;
}


void
StartUpload(PZD1205_ADAPTER Adapter, 
			ULONG		PhyAddr
			)
{
#ifdef HOST_IF_USB
	return;
#endif
	if (!Adapter->DeviceInSleep){
		ZD1205_WRITE_REGISTER(Adapter,PCI_RxAddr_p2, 0);
		ZD1205_WRITE_REGISTER(Adapter,PCI_RxAddr_p1, PhyAddr);
	}
}

BOOLEAN
ProcessWakeupInterrupt(
				IN PZD1205_ADAPTER	Adapter
				   )
{
    PZD1205SwTcb		SwTcb;
    PZD1205SwTcb		NextSwTcb;
    PZD1205SwRfd		SwRfd;
    PZD1205SwRfd		NextSwRfd;
	ULONG				tmpvalue;
	ULONG				tmpvalue1;
	ULONG				ReadValue;
	USHORT				i;
	ULONGLONG			TSFTimer;
	ZD1211_IRQL			OldIrql;
    
	DEBUGFUNC(_T("ProcessWakeupInterrupt"));

	//++ Turn on RF
	TurnOn_RF(Adapter);
	//--

	ZD1205_READ_REGISTER(Adapter,TSF_LowPart, &tmpvalue);
	TSFTimer = tmpvalue;
	ZD1205_READ_REGISTER(Adapter,TSF_HighPart, &tmpvalue);
	TSFTimer += (((ULONGLONG)tmpvalue) << 32);
	TSFTimer = TSFTimer >> 10; // in unit of TU
	INITSTR((_T("TSF(TU) %d \n"), TSFTimer));
	INITSTR((_T("BeaconInterval = %d\n"), Adapter->CardSetting.BeaconInterval));
	INITSTR((_T("TSF mod BeaconInterval = %d\n"), (TSFTimer % Adapter->CardSetting.BeaconInterval)));
	INITSTR((_T("Now, Device had been waken up\n")));
	ZD1205_READ_REGISTER(Adapter,DeviceState, &tmpvalue);
	INITSTR((_T("DeviceState == %x\n"), tmpvalue));
	// In IBSS mode, BCNATIM is now operating, therefore, the Tx-State will not
	// stay in idle state. So, we change form 0xffff to 0xff, ie, we just make
	// sure that bus-masters, both Tx and Rx, are in idle-state.
	//ASSERT((tmpvalue & 0xffff) == 0);
	ASSERT((tmpvalue & 0xff) == 0);
	if (Adapter->TestFlag & PS_DBG){
		ZD1205_WRITE_REGISTER(Adapter,0x804, 0x200);
		ZD1205_READ_REGISTER(Adapter,0x808, &tmpvalue1);
		ZDDbgPrint((_T("\n")));
		for (i=0;i<tmpvalue1;i++){
			ZD1205_READ_REGISTER(Adapter,0x800, &ReadValue);
			ZDDbgPrint((_T("%8x  "), ReadValue));
		}
		ZDDbgPrint((_T("\n")));
	}


	Adapter->DeviceInSleep = FALSE;
	Adapter->SleepProtect = FALSE;

	// Solve Sequence number duplication problem after wakeup.
	Adapter->SequenceNum = 0;

#ifdef HOST_IF_USB
	// In PCI it run in spinlock, but USB not.
    //KeAcquireSpinLock(&Adapter->Lock, &OldIrql);
	ZDDbg((_T("Get SpinLock\n")));
	ZD1211_AcquireSpinLock(Adapter, &Adapter->Lock, &OldIrql);
#endif

	SwTcb = (PZD1205SwTcb) QueueGetHead(&Adapter->ActiveChainList);
	if (SwTcb){
		//++ Make sure TCBs are in a clear state
		SwTcb->Tcb->CbStatus = 0xffffffff;
		NextSwTcb = SwTcb;
		while(1){
			NextSwTcb = (PZD1205SwTcb)QueueGetNext(&Adapter->ActiveChainList, 
																&NextSwTcb->Link);
			if (NextSwTcb){
				NextSwTcb->Tcb->CbStatus = 0xffffffff;
			}
			else{
				break;
			}
		}
		//--
		StartDownload(Adapter, SwTcb->TcbPhys);
	}
	else{
		ProcessQueuedPackets(Adapter);
	}

    SwRfd = (PZD1205SwRfd) QueueGetHead(&Adapter->RfdList);
	if (SwRfd){
		//++ Make sure Rfds are in a clear state
		SwRfd->Rfd->CbStatus = 0xffffffff;
		NextSwRfd = SwRfd;
		while(1){
			NextSwRfd = (PZD1205SwRfd)QueueGetNext(&Adapter->RfdList, 
																&NextSwRfd->Link);
			if (NextSwRfd){
				NextSwRfd->Rfd->CbStatus = 0xffffffff;
			}
			else{
				break;
			}
		}
		//--

		StartUpload(Adapter, SwRfd->RfdPhys);
	}

#ifdef HOST_IF_USB
	//KeReleaseSpinLock(&Adapter->Lock, OldIrql);
	ZDDbg((_T("---------->>> Release SpinLock\n")));
	ZD1211_ReleaseSpinLock(Adapter, &Adapter->Lock, OldIrql);
#endif

	return TRUE;
}

BOOLEAN
ProcessBusAbortInterrupt(
                   IN PZD1205_ADAPTER Adapter
                   )                 
{
    DEBUGFUNC(_T("ProcessBusAbortInterrupt"));

	ZDDbgPrint((_T("################################################\n")));
	ZDDbgPrint((_T(" Bus Abort \n")));
	ZDDbgPrint((_T("################################################\n")));

	return TRUE;
}

UCHAR Rate_Adaption(IN PZD1205_ADAPTER Adapter, 
					PUCHAR	pMacAddr, 
					UCHAR   CurrentRate, 
					UCHAR   gear
					)
{
	UCHAR	i;
	UCHAR	NewRate;

	if (gear == FALL_RATE){
		if (CurrentRate >= 1){
			NewRate = CurrentRate-1;
		}
		else{
			NewRate = CurrentRate;
		}

		for (i=0;i<Adapter->OppositeCnt;i++){
			if (!memcmp(pMacAddr, Adapter->OppositeRateInfo[i].MacAddr, 6) ){
			    
				Adapter->OppositeRateInfo[i].FailFrames++;

				break;
			}
		}
		ZDDbgPrint((_T("MAC1: %x-%x-%x-%x-%x-%x\n"),
			pMacAddr[0],
			pMacAddr[1],
			pMacAddr[2],
			pMacAddr[3],
			pMacAddr[4],
			pMacAddr[5]
					));
		ZDDbgPrint((_T("###### Fall Rate -> %d\n"), NewRate));
		return (NewRate);
	}
	else{
		for (i=0;i<Adapter->OppositeCnt;i++){
			if (!memcmp(pMacAddr, Adapter->OppositeRateInfo[i].MacAddr, 6) ){
				if ((CurrentRate + 1) <= Adapter->OppositeRateInfo[i].MaxRate){
					NewRate = CurrentRate+1;
#if	(!(defined(GCCK) && defined(OFDM)))				
					// 5.5M -> 11M(x) -> 16.5M
					// 16.5M -> 22M(x) -> 27.5M
				    if ( ((NewRate != Adapter->OppositeRateInfo[i].MaxRate)&&(NewRate == 3))
						|| (NewRate == 5) ){
                        NewRate = NewRate + 1;
					}
#endif
					Adapter->OppositeRateInfo[i].CurrentRate = NewRate;
				}
				break;
			}
		}
		if (i == Adapter->OppositeCnt){
			if ((CurrentRate +1) <= CCK_11M){
				NewRate = CurrentRate + 1;
			}
			else{
				NewRate = CCK_11M;
			}
		}
		ZDDbgPrint((_T("MAC1: %x-%x-%x-%x-%x-%x\n"),
			pMacAddr[0],
			pMacAddr[1],
			pMacAddr[2],
			pMacAddr[3],
			pMacAddr[4],
			pMacAddr[5]
					));
		ZDDbgPrint((_T("###### Rise Rate -> %d\n"), NewRate));
		return (NewRate);
	}
}

UCHAR Rate_Convert_to_MS(UCHAR rate_in_drv)
{

#ifdef OFDM
	switch (rate_in_drv)
	{ 
	    case 0  :	return 2;  // 1M
	    case 1  :	return 4;  // 2M
	    case 2 :	return 11;  // 5.5M
	    case 3 :	return 22;  // 11M
	    case 4 :	return 12;  // 6M
	    case 5 :	return 18;  // 9M
	    case 6 :	return 24;  // 12M
	    case 7 :	return 36;	// 18M
	    case 8 :	return 48;	// 24M
	    case 9 :	return 72;	// 36M
	    case 0xa :	return 96;	// 48M
	    case 0xb:	return 108;	// 54M
	}
#endif
	return 0;
}

UCHAR Rate_Convert(UCHAR rate, USHORT capability)
{

#ifdef OFDM
	switch (rate)
	{ 
	    case 2  :	return 0;  // 1M
	    case 4  :	return 1;  // 2M
	    case 11 :	return 2;  // 5.5M
	    case 22 :	return 3;  // 11M
	    case 12 :	return 4;  // 6M
	    case 18 :	return 5;  // 9M
	    case 24 :	return 6;  // 12M
	    case 36 :	return 7;	// 18M
	    case 48 :	return 8;	// 24M
	    case 72 :	return 9;	// 36M
	    case 96 :	return 0xa;	// 48M
	    case 108:	return 0xb;	// 54M
	}
#else
	switch (rate)
	{ 
	    case 2  :  return 0;  // 1M
	    case 4  :  return 1;  // 2M
	    case 11 :  return 2;  // 5.5M
	    case 22 :  return 3;  // 11M
	    case 33 :  return 4;  // 16.5M
	    case 44 :  return 5;  // 22M
	    case 55 :  return 6;  // 27.5M
#if defined(ECCK_60_5)
	    case 66 :  return 7;  // 33M
	    case 77 :  return 8;  // 38.5M
	    case 88 :  return 9;  // 44M
	    case 99 :  return 10; // 49.5M
	    case 110:  return 11; // 55M
	    case 121:  return 12; // 60.5M
#endif
	}
#endif

	return 100;	// 11M
}
void
ProcessQueuedPackets(
                   IN PZD1205_ADAPTER Adapter
					)
{
    NDIS_STATUS         SendStatus;

    while ((Adapter->FirstTxQueue) &&
        ((PZD1205SwTcb) QueueGetHead(&Adapter->TxCBList)))
    {
        PNDIS_PACKET QueuePacket;
		//ASSERT(0);
        // If any packets are in the queue, dequeue it, send it, and
        // acknowledge success.
        QueuePacket = Adapter->FirstTxQueue;
        Adapter->NumPacketsQueued--;
        DequeuePacket(Adapter->FirstTxQueue, Adapter->LastTxQueue);

        // Attempt to put it onto the hardware
        SendStatus = SetupNextSend(Adapter, QueuePacket, DATA, 
							Adapter->CardSetting.EncryMode, UPPER_NDIS, NULL);

        // If there were no resources for this packet, then we'll just have
        // to try and send it later.
        if (SendStatus == NDIS_STATUS_RESOURCES)
        {
            // re-queue the packet
            Adapter->NumPacketsQueued++;
            EnqueuePacket(Adapter->FirstTxQueue, Adapter->LastTxQueue, QueuePacket);
            break;
        }
    }
}
